import concurrent.futures
import json
import re
import threading
import time
from collections.abc import Iterable, Iterator
from datetime import date, datetime, timezone
from io import TextIOWrapper
from itertools import product
from pathlib import Path
from typing import Any

import pytest

from cognite_toolkit._cdf_tk.exceptions import ToolkitError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.fileio import (
    COMPRESSION_BY_NAME,
    COMPRESSION_BY_SUFFIX,
    FILE_READ_CLS_BY_FORMAT,
    FILE_WRITE_CLS_BY_FORMAT,
    Chunk,
    Compression,
    CSVReader,
    CSVWriter,
    FailedParsing,
    FileReader,
    FileWriter,
    SchemaColumn,
    Uncompressed,
)
from cognite_toolkit._cdf_tk.utils.fileio._readers import YAMLBaseReader
from cognite_toolkit._cdf_tk.utils.fileio._writers import NDJsonWriter, YAMLBaseWriter
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


@pytest.fixture()
def json_chunks() -> tuple[list[dict[str, JsonVal]], list[SchemaColumn]]:
    chunks = [
        {"text": "value1", "integer": 123},
        {"text": "value4", "list": [1, 2, 3], "nested": {"key": "value"}},
        {"text": "value5", "boolean": True},
        {"text": "value6"},
        {"text": "value7", "float": 3.14, "empty_list": []},
    ]
    schema = [
        SchemaColumn(name="text", type="string"),
        SchemaColumn(name="integer", type="integer"),
        SchemaColumn(name="list", type="integer", is_array=True),
        SchemaColumn(name="nested", type="json"),
        SchemaColumn(name="boolean", type="boolean"),
        SchemaColumn(name="float", type="float"),
        SchemaColumn(name="empty_list", type="float", is_array=True),
    ]
    return chunks, schema


@pytest.fixture()
def cell_chunks(
    json_chunks: tuple[list[dict[str, JsonVal]], list[SchemaColumn]],
) -> tuple[list[Chunk], list[SchemaColumn]]:
    chunks, schema = json_chunks
    chunks = [
        *chunks,
        {"date": date(2023, 10, 1), "timestamp": datetime(2023, 10, 1, 12, 0, 0, tzinfo=timezone.utc)},
    ]
    schema = [
        *schema,
        SchemaColumn(name="date", type="date"),
        SchemaColumn(name="timestamp", type="timestamp"),
    ]
    return chunks, schema


class LineReader(FileReader):
    format = ".DummyFormat"

    def _read_chunks_from_file(self, file: TextIOWrapper) -> Iterator[JsonVal]:
        for line in file:
            if line.strip():
                yield {"line": line.strip()}


class DummyWriter(FileWriter[TextIOWrapper]):
    format = ".dummy"

    def __init__(self, output_dir: Path) -> None:
        super().__init__(output_dir=output_dir, kind="DummyKind", compression=Uncompressed)
        self.written_chunks: list[Chunk] = []
        self.opened_files: list[Path] = []

    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        self.opened_files.append(filepath)
        writer = self.compression_cls(filepath).open(mode="w")
        return writer

    def _write(self, writer: TextIOWrapper, chunks: Iterable[Chunk]) -> None:
        self.written_chunks.extend(chunks)
        writer.write("Writing chunks\n")


class TestCompression:
    @pytest.mark.parametrize(
        "compression_name",
        list(COMPRESSION_BY_NAME.keys()),
    )
    def test_read_write_compression_by_name(self, compression_name: str, tmp_path: Path) -> None:
        compression = Compression.from_name(compression_name)
        tmp_path = tmp_path / f"test_file.txt{compression.file_suffix}"
        with compression(tmp_path).open(mode="w") as file:
            file.write("Test content")

        with compression(tmp_path).open("r") as file:
            content = file.read()

        assert content == "Test content"

    @pytest.mark.parametrize(
        "compression_suffix",
        [*COMPRESSION_BY_SUFFIX.keys(), Uncompressed.file_suffix],
    )
    def test_read_write_compression_by_suffix(self, compression_suffix: str, tmp_path: Path) -> None:
        tmp_path = tmp_path / f"test_file.txt{compression_suffix}"
        with Compression.from_filepath(tmp_path).open(mode="w") as file:
            file.write("Test content")

        with Compression.from_filepath(tmp_path).open("r") as file:
            content = file.read()

        assert content == "Test content"

    def test_all_compression_classes_registered(self) -> None:
        expected_compressions = get_concrete_subclasses(Compression)

        assert set(COMPRESSION_BY_NAME.values()) == set(expected_compressions)
        assert set(COMPRESSION_BY_SUFFIX.values()) == set(expected_compressions)


class TestFileWriter:
    def test_file_split_on_limit(self, tmp_path: Path) -> None:
        dummy_writer = DummyWriter(tmp_path / "dummy")
        dummy_writer.max_file_size_bytes = 1  # Set a small limit for testing
        chunk1 = [{"a": 1}]
        chunk2 = [{"b": 2}]
        with dummy_writer:
            dummy_writer.write_chunks(chunk1, filestem="splitfile")
            dummy_writer.write_chunks(chunk2, filestem="splitfile")
        assert [file.relative_to(tmp_path) for file in dummy_writer.opened_files] == [
            Path("dummy/splitfile-part-0000.DummyKind.dummy"),
            Path("dummy/splitfile-part-0001.DummyKind.dummy"),
        ]
        assert dummy_writer.written_chunks == chunk1 + chunk2

    def test_multiple_filenames(self, tmp_path: Path) -> None:
        output_dir = tmp_path / "dummy"
        dummy_writer = DummyWriter(output_dir)
        chunk1 = [{"x": 1}]
        chunk2 = [{"y": 2}]
        with dummy_writer:
            dummy_writer.write_chunks(chunk1, filestem="file1")
            dummy_writer.write_chunks(chunk2, filestem="file2")
        assert [file.relative_to(tmp_path) for file in dummy_writer.opened_files] == [
            Path("dummy/file1-part-0000.DummyKind.dummy"),
            Path("dummy/file2-part-0000.DummyKind.dummy"),
        ]

    def test_all_file_writers_registered(self) -> None:
        # YAMLBaseWriter is an abstract class, so we exclude it from the expected readers
        expected_writers = set(get_concrete_subclasses(FileWriter)) - {DummyWriter, YAMLBaseWriter}

        assert set(FILE_WRITE_CLS_BY_FORMAT.values()) == expected_writers

    def test_create_from_format_raises(self) -> None:
        with pytest.raises(ToolkitValueError) as excinfo:
            FileWriter.create_from_format("unknown_format", Path("."), "DummyKind", Uncompressed)
        assert str(excinfo.value).startswith("Unknown file format: unknown_format. Available formats: ")


class TestFileWriterThreadSafety:
    """Test thread safety of FileWriter implementations."""

    def test_concurrent_write_chunks_same_filestem(self, tmp_path: Path) -> None:
        """Test that concurrent writes to the same filestem are thread-safe."""
        output_dir = tmp_path
        writer = NDJsonWriter(output_dir, "test", Uncompressed)

        # Prepare test data
        num_threads = 10
        chunks_per_thread = 100
        all_chunks = []

        def write_worker(thread_id: int) -> list[dict[str, Any]]:
            """Worker function that writes chunks and returns what it wrote."""
            chunks = [
                {"thread": thread_id, "chunk": i, "data": f"test_data_{thread_id}_{i}"}
                for i in range(chunks_per_thread)
            ]
            writer.write_chunks(chunks, "test_file")
            return chunks

        # Execute concurrent writes
        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor, writer:
            futures = [executor.submit(write_worker, i) for i in range(num_threads)]
            for future in concurrent.futures.as_completed(futures):
                all_chunks.extend(future.result())

        # Verify all data was written correctly
        written_files = list(output_dir.glob("*.ndjson"))
        assert len(written_files) > 0

        written_data = []
        for file_path in written_files:
            with file_path.open("r", encoding=Compression.encoding) as file:
                for line in file:
                    written_data.append(json.loads(line.strip()))

        # Verify all chunks were written
        assert len(written_data) == len(all_chunks)

        # Verify data integrity - each chunk should appear exactly once
        written_keys = {(chunk["thread"], chunk["chunk"]) for chunk in written_data}
        expected_keys = {(chunk["thread"], chunk["chunk"]) for chunk in all_chunks}
        assert written_keys == expected_keys

    def test_concurrent_write_chunks_different_filestems(self, tmp_path: Path) -> None:
        """Test concurrent writes to different filestems are properly isolated."""
        output_dir = tmp_path
        writer = NDJsonWriter(output_dir, "test", Uncompressed)

        num_threads = 5
        chunks_per_thread = 50

        def write_worker(thread_id: int) -> str:
            """Worker function that writes to a unique filestem."""
            filestem = f"file_{thread_id}"
            chunks = [{"thread": thread_id, "chunk": i} for i in range(chunks_per_thread)]
            writer.write_chunks(chunks, filestem)
            return filestem

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor, writer:
            futures = [executor.submit(write_worker, i) for i in range(num_threads)]
            _ = [future.result() for future in concurrent.futures.as_completed(futures)]

        # Verify each filestem has its own files
        for thread_id in range(num_threads):
            thread_files = list(output_dir.glob(f"file_{thread_id}-*.ndjson"))
            assert len(thread_files) > 0

            # Verify data in thread-specific files
            written_data = []
            for file_path in thread_files:
                with file_path.open("r", encoding=Compression.encoding) as file:
                    for line in file:
                        data = json.loads(line.strip())
                        assert data["thread"] == thread_id
                        written_data.append(data)

            assert len(written_data) == chunks_per_thread

    def test_concurrent_file_size_limit_handling(self, tmp_path):
        """Test that file size limit handling is thread-safe."""
        output_dir = tmp_path
        # Use very small file size limit to force file rotation
        writer = NDJsonWriter(output_dir, "test", Uncompressed, max_file_size_bytes=1024)

        num_threads = 5
        large_chunks_per_thread = 20

        def write_worker(thread_id: int) -> None:
            """Worker that writes large chunks to trigger file rotation."""
            for i in range(large_chunks_per_thread):
                # Create a large chunk to exceed file size limit
                large_data = "x" * 200  # Make chunk large enough
                chunk = {"thread": thread_id, "chunk": i, "large_data": large_data}
                writer.write_chunks([chunk], "large_file")

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor, writer:
            futures = [executor.submit(write_worker, i) for i in range(num_threads)]
            for future in concurrent.futures.as_completed(futures):
                future.result()  # Wait for completion

        # Verify multiple files were created due to size limits
        written_files = list(output_dir.glob("large_file-*.ndjson"))
        assert len(written_files) > 1  # Should have multiple files due to size limit

        # Verify all data is present across files
        total_chunks = 0
        thread_counts = {}
        for file_path in written_files:
            with open(file_path, encoding=Compression.encoding) as f:
                for line in f:
                    data = json.loads(line.strip())
                    total_chunks += 1
                    thread_id = data["thread"]
                    thread_counts[thread_id] = thread_counts.get(thread_id, 0) + 1

        expected_total = num_threads * large_chunks_per_thread
        assert total_chunks == expected_total

        # Verify each thread's data is complete
        for thread_id in range(num_threads):
            assert thread_counts.get(thread_id, 0) == large_chunks_per_thread

    def test_csv_writer_thread_safety(self, tmp_path):
        """Test thread safety specifically for CSVWriter with its internal state."""
        output_dir = tmp_path
        columns = [
            SchemaColumn("id", "integer", False),
            SchemaColumn("name", "string", False),
            SchemaColumn("value", "float", False),
        ]
        writer = CSVWriter(output_dir, "test", Uncompressed, columns)

        num_threads = 8
        rows_per_thread = 25

        def csv_write_worker(thread_id: int) -> None:
            """Worker that writes CSV data."""
            chunks = [
                {"id": thread_id * 1000 + i, "name": f"name_{thread_id}_{i}", "value": float(i)}
                for i in range(rows_per_thread)
            ]
            writer.write_chunks(chunks, f"csv_test_{thread_id}")

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor, writer:
            futures = [executor.submit(csv_write_worker, i) for i in range(num_threads)]
            for future in concurrent.futures.as_completed(futures):
                future.result()

        # Verify CSV files were created and have correct structure
        csv_files = list(output_dir.glob("*.csv"))
        assert len(csv_files) > 0

        total_rows = 0
        for csv_file in csv_files:
            with open(csv_file) as f:
                lines = f.readlines()
                # First line should be header
                assert "id,name,value" in lines[0]
                # Count data rows (excluding header)
                total_rows += len(lines) - 1

        expected_rows = num_threads * rows_per_thread
        assert total_rows == expected_rows

    def test_race_condition_file_count_tracking(self, tmp_path):
        """Test that file count tracking doesn't have race conditions."""
        output_dir = tmp_path
        writer = NDJsonWriter(output_dir, "test", Uncompressed, max_file_size_bytes=512)

        # Track file counts from multiple threads
        file_counts = []
        lock = threading.Lock()

        def count_tracking_worker(thread_id: int) -> None:
            """Worker that tracks file counts during concurrent writes."""
            for i in range(20):
                # Write data that might trigger file rotation
                chunk = {"thread": thread_id, "index": i, "data": "x" * 100}
                writer.write_chunks([chunk], "count_test")

                # Record file count
                with lock:
                    count = writer._file_count_by_filename["count_test"]
                    file_counts.append(count)

                # Small delay to increase chance of race conditions
                time.sleep(0.001)

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor, writer:
            futures = [executor.submit(count_tracking_worker, i) for i in range(5)]
            for future in concurrent.futures.as_completed(futures):
                future.result()

        # File counts should be monotonically increasing (never decrease)
        assert file_counts == sorted(file_counts)

        # Final file count should match actual files created
        actual_files = list(output_dir.glob("*.ndjson"))
        expected_file_stems = [f"count_test-part-{count:04}.test" for count in sorted(set(file_counts))]
        actual_file_stems = sorted([file.stem for file in actual_files])
        # Note: writer.file_count might be 0 after context exit as it clears state
        assert actual_file_stems == expected_file_stems

    @pytest.mark.parametrize(
        "num_threads,chunks_per_thread",
        [
            (2, 10),  # Light load
            (5, 50),  # Medium load
            (10, 100),  # Heavy load
        ],
    )
    def test_stress_concurrent_writes(self, tmp_path: Path, num_threads: int, chunks_per_thread: int):
        """Stress test concurrent writes with varying loads."""
        output_dir = tmp_path
        writer = NDJsonWriter(output_dir, "stress_test", Uncompressed)

        def stress_worker(thread_id: int) -> int:
            """Worker that performs stress writes."""
            chunks_written = 0
            for batch in range(5):  # Multiple batches per thread
                chunks = [
                    {"thread": thread_id, "batch": batch, "chunk": i, "timestamp": time.time()}
                    for i in range(chunks_per_thread // 5)
                ]
                writer.write_chunks(chunks, f"stress_{thread_id}")
                chunks_written += len(chunks)
            return chunks_written

        start_time = time.time()

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor, writer:
            futures = [executor.submit(stress_worker, i) for i in range(num_threads)]
            total_written = sum(future.result() for future in concurrent.futures.as_completed(futures))

        end_time = time.time()

        # Verify performance is reasonable (should complete within reasonable time)
        duration = end_time - start_time
        assert duration < 30.0, f"Stress test took too long: {duration}s"

        # Verify all data was written
        files = list(output_dir.glob("*.ndjson"))
        assert len(files) > 0

        written_count = 0
        for file_path in files:
            with open(file_path) as f:
                written_count += sum(1 for _ in f)

        assert written_count == total_written


class TestFileReader:
    def test_read_multiple_lines(self, tmp_path: Path) -> None:
        # Create a dummy file with multiple lines
        file_content = "line1\nline2\nline3\n"
        file_path = tmp_path / "dummy.txt"
        file_path.write_text(file_content, encoding="utf-8")
        reader = LineReader(file_path)
        chunks = list(reader.read_chunks())
        assert chunks == [
            {"line": "line1"},
            {"line": "line2"},
            {"line": "line3"},
        ]

    def test_read_empty_file(self, tmp_path: Path) -> None:
        file_path = tmp_path / "empty.txt"
        file_path.touch()
        reader = LineReader(file_path)
        chunks = list(reader.read_chunks())
        assert chunks == []

    def test_read_lines_with_whitespace(self, tmp_path: Path) -> None:
        file_content = "  line1  \n\n  line2\n   \nline3   \n"
        file_path = tmp_path / "whitespace.txt"
        file_path.write_text(file_content, encoding="utf-8")
        reader = LineReader(file_path)
        chunks = list(reader.read_chunks())
        assert chunks == [
            {"line": "line1"},
            {"line": "line2"},
            {"line": "line3"},
        ]

    def test_all_file_readers_registered(self) -> None:
        # YAMLBaseReader is an abstract class, so we exclude it from the expected readers
        expected_readers = set(get_concrete_subclasses(FileReader)) - {LineReader, YAMLBaseReader}

        assert set(expected_readers) == set(FILE_READ_CLS_BY_FORMAT.values())

    @pytest.mark.parametrize(
        "filepath, expected_error",
        [
            (Path("non_existent_file.txt"), "Unknown file format: .txt. Available formats:"),
            (Path("file_without_format"), "File has no suffix. Available formats:"),
            (Path("file_with_multiple_suffixes.txt.gz"), "Unknown file format: .txt."),
            (Path("file_with_only_compression.gz"), "File has a compression suffix, but no file format suffix found."),
        ],
    )
    def test_from_filepath_raise(self, filepath: Path, expected_error: str) -> None:
        with pytest.raises(ToolkitValueError) as excinfo:
            FileReader.from_filepath(filepath)

        assert str(excinfo.value).startswith(expected_error)


class TestFileIO:
    @pytest.mark.parametrize(
        "format, compression_name",
        list(product(FILE_WRITE_CLS_BY_FORMAT.keys(), COMPRESSION_BY_NAME.keys())),
    )
    def test_write_read(
        self,
        format: str,
        compression_name: str,
        json_chunks: tuple[list[dict[str, JsonVal]], list[SchemaColumn]],
        tmp_path: Path,
    ) -> None:
        chunks, columns = json_chunks
        chunk_len = len(chunks)
        compression_cls = COMPRESSION_BY_NAME[compression_name]
        output_dir = tmp_path / "output"
        with FileWriter.create_from_format(
            format, output_dir, "Test", compression=compression_cls, columns=columns
        ) as writer:
            mid = chunk_len // 2
            writer.write_chunks(chunks[:mid])
            writer.write_chunks(chunks[mid:])

        file_path = list(output_dir.rglob(f"*{format}{compression_cls.file_suffix}"))
        assert len(file_path) == 1

        reader = FileReader.from_filepath(file_path[0])
        # The Table reader returns all columns, even if they are None, so we filter them out for comparison.
        read_chunks = [
            {key: value for key, value in chunk.items() if value is not None} for chunk in reader.read_chunks()
        ]

        assert read_chunks == chunks

    @pytest.mark.parametrize(
        "format, compression_name",
        list(product(FILE_WRITE_CLS_BY_FORMAT.keys(), COMPRESSION_BY_NAME.keys())),
    )
    def test_write_cell_chunks(
        self, format: str, compression_name: str, cell_chunks: tuple[list[Chunk], list[SchemaColumn]], tmp_path: Path
    ) -> None:
        chunks, columns = cell_chunks
        compression_cls = COMPRESSION_BY_NAME[compression_name]
        output_dir = tmp_path / "output"
        with FileWriter.create_from_format(
            format, output_dir, "Test", compression=compression_cls, columns=columns
        ) as writer:
            writer.write_chunks(chunks)

        file_path = list(output_dir.rglob(f"*{format}{compression_cls.file_suffix}"))
        assert len(file_path) == 1
        # We cannot read cell chunks directly, as these contain non-JSON values like dates and timestamps.

    @pytest.mark.parametrize(
        "format, compression_name",
        list(product(FILE_WRITE_CLS_BY_FORMAT.keys(), COMPRESSION_BY_NAME.keys())),
    )
    def test_write_split_files(
        self,
        format: str,
        compression_name: str,
        json_chunks: tuple[list[dict[str, JsonVal]], list[SchemaColumn]],
        tmp_path: Path,
    ) -> None:
        chunks, columns = json_chunks
        chunk_len = len(chunks)
        compression_cls = COMPRESSION_BY_NAME[compression_name]
        output_dir = tmp_path / "output"
        writer_inst = FileWriter.create_from_format(
            format, output_dir, "Test", compression=compression_cls, columns=columns
        )
        writer_inst.max_file_size_bytes = 1  # Small size to force splitting
        with writer_inst as writer:
            mid = chunk_len // 2
            writer.write_chunks(chunks[:mid])
            writer.write_chunks(chunks[mid:])

        def part_number(filepath: Path) -> int:
            match = re.match(r"part-(\d{4}).*", filepath.name)
            return int(match.group(1)) if match else -1

        file_path = sorted(output_dir.rglob(f"*{format}{compression_cls.file_suffix}"), key=part_number)
        assert len(file_path) == 2

        reader = FileReader.from_filepath(file_path[0])
        # The Table reader returns all columns, even if they are None, so we filter them out for comparison.
        read_chunks = [
            {key: value for key, value in chunk.items() if value is not None} for chunk in reader.read_chunks()
        ]
        assert read_chunks == chunks[:mid]
        reader = FileReader.from_filepath(file_path[1])
        read_chunks = [
            {key: value for key, value in chunk.items() if value is not None} for chunk in reader.read_chunks()
        ]
        assert read_chunks == chunks[mid:]


class TestCSVReader:
    CSV_CONTENT = """text,integer,nested,boolean,float
value1,123,"{""key"": ""value""}",true,3.14
value2,456,"{""key"": ""value2""}",false,2.71
,,,,
value3,789,"{""key"": ""value3""}",true,1.41
310,false,31.2,text,20
"""
    EXPECTED_SCHEMA = (
        SchemaColumn(name="text", type="string"),
        SchemaColumn(name="integer", type="integer"),
        SchemaColumn(name="nested", type="json"),
        SchemaColumn(name="boolean", type="boolean"),
        SchemaColumn(name="float", type="float"),
    )

    def test_sniff_schema(self, tmp_path: Path) -> None:
        csv_path = tmp_path / "test.csv"
        csv_path.write_text(self.CSV_CONTENT, encoding="utf-8")

        schema = CSVReader.sniff_schema(csv_path, sniff_rows=100)

        assert schema == list(self.EXPECTED_SCHEMA)

    @pytest.mark.parametrize(
        "content,filename,expected_error",
        [
            pytest.param("", "my_file.csv", "No data found in the file: '{filepath}'.", id="empty file"),
            pytest.param("", None, "File not found: '{filepath}'.", id="missing file"),
            pytest.param(
                "some random text",
                "invalid.txt",
                "Expected a .csv file got a '.txt' file instead.",
                id="invalid format",
            ),
            pytest.param(
                "header1,header1\nvalue1,value2",
                "dup_headers.csv",
                "Duplicate column names found: header1.",
                id="duplicate headers",
            ),
            pytest.param(
                "header1,header2\n", "no_data.csv", "No data found in the file: '{filepath}'.", id="no data rows"
            ),
        ],
    )
    def test_sniff_schema_error_cases(
        self, content: str, filename: str | None, expected_error: str, tmp_path: Path
    ) -> None:
        csv_path = tmp_path / (filename or "missing.csv")
        if filename is not None:
            csv_path.write_text(content, encoding="utf-8")
        if "{filepath}" in expected_error:
            expected_error = expected_error.format(filepath=csv_path.as_posix())

        with pytest.raises(ToolkitError) as excinfo:
            CSVReader.sniff_schema(csv_path, sniff_rows=100)

        assert str(excinfo.value) == expected_error

    def test_read_with_schema(self, tmp_path: Path) -> None:
        csv_path = tmp_path / "test.csv"
        csv_path.write_text(self.CSV_CONTENT, encoding="utf-8")

        reader = CSVReader(csv_path, schema=self.EXPECTED_SCHEMA, keep_failed_cells=True)
        chunks = list(reader.read_chunks())

        assert len(chunks) == 5
        assert chunks == [
            {
                "text": "value1",
                "integer": 123,
                "nested": {"key": "value"},
                "boolean": True,
                "float": 3.14,
            },
            {
                "text": "value2",
                "integer": 456,
                "nested": {"key": "value2"},
                "boolean": False,
                "float": 2.71,
            },
            {"boolean": None, "float": None, "integer": None, "nested": None, "text": None},
            {"boolean": True, "float": 1.41, "integer": 789, "nested": {"key": "value3"}, "text": "value3"},
            {"boolean": None, "float": 20.0, "integer": None, "nested": 31.2, "text": "310"},
        ]
        assert len(reader.failed_cell) == 2
        assert reader.failed_cell == [
            FailedParsing(row=5, column="integer", value="false", error="Cannot convert false to int64."),
            FailedParsing(row=5, column="boolean", value="text", error="Cannot convert text to boolean."),
        ]

    def test_read_unprocessed_csv(self, tmp_path: Path) -> None:
        csv_content = "id,space,externalId,number\n1,space1,id1,1.30\n2,space2,id2,42.0\n"
        csv_file = tmp_path / "test.csv"
        csv_file.write_text(csv_content, encoding="utf-8")

        chunks = list(CSVReader(csv_file).read_chunks_unprocessed())

        assert chunks == [
            {"id": "1", "space": "space1", "externalId": "id1", "number": "1.30"},
            {"id": "2", "space": "space2", "externalId": "id2", "number": "42.0"},
        ]



from abc import ABC
from dataclasses import dataclass
from datetime import date, datetime
from typing import IO, ClassVar, TypeAlias, TypeVar

from cognite_toolkit._cdf_tk.utils.useful_types import DataType, JsonVal

PrimaryCellValue: TypeAlias = datetime | date | JsonVal
CellValue: TypeAlias = PrimaryCellValue | list[PrimaryCellValue]
Chunk: TypeAlias = dict[str, CellValue]


T_IO = TypeVar("T_IO", bound=IO)


class FileIO(ABC):
    format: ClassVar[str]


@dataclass(frozen=True)
class SchemaColumn:
    name: str
    type: DataType
    is_array: bool = False

    def __post_init__(self) -> None:
        if self.type == "json" and self.is_array:
            raise ValueError("JSON columns cannot be arrays. Use 'is_array=False' for JSON columns.")



from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.cruds import RawFileCRUD, RawTableCRUD
from cognite_toolkit._cdf_tk.data_classes import BuildEnvironment, BuiltResource, BuiltResourceList, SourceLocationEager


class TestRawFileLoader:
    @pytest.mark.parametrize(
        "csv_content, expected_write",
        [
            (
                """myFloat,myInt,myString,myBool
0.1,1,hello,True
0.2,2,world,False
""",
                {
                    0: {"myFloat": 0.1, "myInt": 1, "myString": "hello", "myBool": True},
                    1: {"myFloat": 0.2, "myInt": 2, "myString": "world", "myBool": False},
                },
            )
        ],
    )
    def test_upload_dtypes(self, csv_content: str, expected_write: dict[int, Any]) -> None:
        with monkeypatch_toolkit_client() as client:
            loader = RawFileCRUD.create_loader(client)
        csv_file = MagicMock(spec=Path)
        csv_file.read_bytes.return_value = csv_content.encode("utf-8")
        csv_file.exists.return_value = True
        csv_file.suffix = ".csv"
        source_file = MagicMock(spec=Path)
        source_file.with_suffix.return_value = csv_file

        state = BuildEnvironment()
        state.built_resources[RawFileCRUD.folder_name] = BuiltResourceList(
            [
                BuiltResource(
                    RawTable("myDB", "myTable"),
                    SourceLocationEager(source_file, "1z234"),
                    RawTableCRUD.kind,
                    None,
                    None,
                )
            ]
        )

        list(loader.upload(state, dry_run=False))

        # Verify one upload call was made
        assert client.raw.rows.insert_dataframe.call_count == 1
        _, kwargs = client.raw.rows.insert_dataframe.call_args
        written_to_cdf = kwargs["dataframe"].to_dict(orient="index")
        # All values and types should match the expected ingestion payload
        assert written_to_cdf == expected_write

    def test_upload_preserves_numeric_types_and_sets_empty_strings_for_nulls(self) -> None:
        with monkeypatch_toolkit_client() as client:
            loader = RawFileCRUD.create_loader(client)
        csv_file = MagicMock(spec=Path)
        csv_content = """myFloat,myInt,myString,myBool
,1,hello,True
0.2,,world,False
"""
        csv_file.read_bytes.return_value = csv_content.encode("utf-8")
        csv_file.exists.return_value = True
        csv_file.suffix = ".csv"
        source_file = MagicMock(spec=Path)
        source_file.with_suffix.return_value = csv_file

        state = BuildEnvironment()
        state.built_resources[RawFileCRUD.folder_name] = BuiltResourceList(
            [
                BuiltResource(
                    RawTable("myDB", "myTable"),
                    SourceLocationEager(source_file, "1z234"),
                    RawTableCRUD.kind,
                    None,
                    None,
                )
            ]
        )

        list(loader.upload(state, dry_run=False))

        # Capture the DataFrame uploaded to RAW

        # Verify one upload call was made
        assert client.raw.rows.insert_dataframe.call_count == 1
        _, kwargs = client.raw.rows.insert_dataframe.call_args
        df = kwargs["dataframe"]
        # Verify dtypes are object after astype(object).fillna("")
        assert str(df.dtypes["myFloat"]) == "object"
        assert str(df.dtypes["myInt"]) == "object"

        # Non-null float value remains numeric
        assert isinstance(df.iloc[1]["myFloat"], float)
        # CSV with nulls coerces an integer-like column to floats; value becomes 1.0
        assert isinstance(df.iloc[0]["myInt"], float)
        # Null in float column becomes empty string
        assert df.iloc[0]["myFloat"] == "" and isinstance(df.iloc[0]["myFloat"], str)
        # Null in int column becomes empty string
        assert df.iloc[1]["myInt"] == "" and isinstance(df.iloc[1]["myInt"], str)



"""
Root conftest.py for shared test fixtures.

Provides worker-scoped fixtures for external library caching to avoid race conditions
and expensive repeated downloads when running tests with pytest-xdist.
"""

import atexit
import hashlib
import os
import shutil
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import pytest
from filelock import FileLock

from cognite_toolkit._cdf_tk.commands.modules import ModulesCommand
from cognite_toolkit._cdf_tk.utils import sanitize_filename

# Global registry to track temp directories for cleanup
_TEMP_DIRS_TO_CLEANUP: set[Path] = set()


def _cleanup_temp_dirs() -> None:
    """Clean up all registered temp directories at process exit."""
    for temp_dir in _TEMP_DIRS_TO_CLEANUP:
        if temp_dir.exists():
            try:
                shutil.rmtree(temp_dir, ignore_errors=True)
            except Exception as e:
                print(f"Warning: Failed to clean up temp directory {temp_dir}: {e}")


atexit.register(_cleanup_temp_dirs)


def get_worker_id() -> str:
    """Get the current pytest-xdist worker ID, or 'master' if not using xdist."""
    worker_id = os.environ.get("PYTEST_XDIST_WORKER", "master")
    return sanitize_filename(worker_id)


@contextmanager
def external_library_cache_context(cache_root: Path) -> Iterator[Path]:
    """Context manager for accessing the external library cache with file locking."""
    worker_id = get_worker_id()
    cache_dir = cache_root / f"modules-{worker_id}"
    lock_file = cache_root / f"modules-{worker_id}.lock"

    cache_dir.mkdir(parents=True, exist_ok=True)
    with FileLock(lock_file, timeout=300):
        yield cache_dir


def get_unique_run_id() -> str:
    """Get a unique identifier for this GitHub Actions job run."""
    run_id = os.getenv("GITHUB_RUN_ID", "")
    run_attempt = os.getenv("GITHUB_RUN_ATTEMPT", "")
    job_id = os.getenv("GITHUB_JOB", "")

    if run_id:
        parts = [run_id]
        if run_attempt:
            parts.append(run_attempt)
        if job_id:
            job_id_safe = sanitize_filename(job_id)
            parts.append(job_id_safe)
        return "-".join(parts)

    import time

    return f"local-{os.getpid()}-{int(time.time() * 1000)}"


@pytest.fixture(scope="session")
def external_library_cache_root() -> Iterator[Path]:
    """Session-scoped fixture providing a root directory for external library caching."""
    run_id = get_unique_run_id()
    session_id = os.getpid()
    cache_root = Path(tempfile.gettempdir()) / f"toolkit-test-cache-{run_id}-{session_id}"
    cache_root.mkdir(parents=True, exist_ok=True)

    _TEMP_DIRS_TO_CLEANUP.add(cache_root)

    yield cache_root

    if cache_root.exists():
        shutil.rmtree(cache_root, ignore_errors=True)


@pytest.fixture(scope="session")
def external_library_modules_dir(external_library_cache_root: Path) -> Iterator[Path]:
    """Session-scoped fixture providing a worker-specific directory for external library downloads."""
    worker_id = get_worker_id()
    modules_dir = external_library_cache_root / f"modules-{worker_id}"
    modules_dir.mkdir(parents=True, exist_ok=True)

    yield modules_dir


@pytest.fixture(scope="function")
def isolated_external_library_dir(external_library_cache_root: Path) -> Iterator[Path]:
    """Function-scoped fixture for tests that need a completely isolated library directory."""
    test_id = os.environ.get("PYTEST_CURRENT_TEST", "unknown")
    test_hash = hashlib.sha256(test_id.encode()).hexdigest()[:8]
    worker_id = get_worker_id()

    isolated_dir = external_library_cache_root / f"isolated-{worker_id}-{test_hash}"
    isolated_dir.mkdir(parents=True, exist_ok=True)

    yield isolated_dir

    if isolated_dir.exists():
        shutil.rmtree(isolated_dir, ignore_errors=True)


@pytest.fixture(scope="function")
def modules_command_with_cached_download(
    external_library_cache_root: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> type:
    """Function-scoped fixture providing a pre-configured ModulesCommand class with caching."""
    worker_id = get_worker_id()
    cache_dir = external_library_cache_root / worker_id / "modules"
    cache_dir.mkdir(parents=True, exist_ok=True)

    lock_file = external_library_cache_root / worker_id / "modules.lock"

    original_init = ModulesCommand.__init__
    original_download = ModulesCommand._download
    original_unpack = ModulesCommand._unpack

    def patched_init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self._temp_download_dir = cache_dir

    def patched_download(self, url: str, file_path: Path) -> None:
        """Download with caching and locking - skip if file already exists."""
        with FileLock(lock_file, timeout=300):
            if file_path.exists():
                if file_path.stat().st_size > 0:
                    return

            unpacked_dir = file_path.parent / file_path.stem
            if unpacked_dir.exists() and unpacked_dir.is_dir():
                try:
                    if any(unpacked_dir.iterdir()):
                        return
                except OSError as e:
                    print(f"Warning: Could not check unpacked directory {unpacked_dir}: {e}")

            original_download(self, url, file_path)

    def patched_unpack(self, file_path: Path) -> None:
        """Unpack with locking to prevent race conditions."""
        with FileLock(lock_file, timeout=300):
            unpacked_dir = file_path.parent / file_path.stem
            if unpacked_dir.exists() and unpacked_dir.is_dir():
                try:
                    if any(unpacked_dir.iterdir()):
                        return
                except OSError as e:
                    print(f"Warning: Could not check unpacked directory {unpacked_dir}: {e}")

            original_unpack(self, file_path)

    def patched_exit(self, *args, **kwargs):
        """Prevent cleanup - cache persists until session end."""
        pass

    monkeypatch.setattr(ModulesCommand, "__init__", patched_init)
    monkeypatch.setattr(ModulesCommand, "_download", patched_download)
    monkeypatch.setattr(ModulesCommand, "_unpack", patched_unpack)
    monkeypatch.setattr(ModulesCommand, "__exit__", patched_exit)

    return ModulesCommand



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, Literal

from cognite.client.data_classes._base import (
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import InstanceApply, NodeId, ViewId
from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel, field_validator, model_validator

from cognite_toolkit._cdf_tk.client.data_classes.instances import InstanceApplyList
from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId
from cognite_toolkit._cdf_tk.commands._migrate.default_mappings import create_default_mappings
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.storageio._data_classes import ModelList
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricKind,
    AssetCentricType,
    JsonVal,
    T_AssetCentricResource,
)


class MigrationMapping(BaseModel, alias_generator=to_camel_case, extra="ignore", populate_by_name=True):
    """The mapping between an asset-centric ID and a data modeling instance ID.
    Args
        resource_type (str): The asset-centric type of the resource (e.g., "asset", "event", "timeseries").
        instance_id (NodeId): The target NodeId in data modeling.
        id (int): The asset-centric ID of the resource.
        data_set_id (int | None): The data set ID of the resource. This is used to validate access to the resource.
        ingestion_view (str | None): The ingestion view name. This is the view mapping that will be used to
            ingest the resource into data modeling.
        preferred_consumer_view (ViewId | None): The preferred consumer view for the resource. This is used in
           for example, the Canvas migration to determine which view to use for the resource.
    """

    resource_type: AssetCentricType
    instance_id: NodeId
    id: int
    data_set_id: int | None = None
    ingestion_view: str | None = None
    preferred_consumer_view: ViewId | None = None

    def get_ingestion_view(self) -> str:
        """Get the ingestion view for the mapping. If not specified, return the default ingestion view."""
        if self.ingestion_view:
            return self.ingestion_view

        default_mappings = create_default_mappings()
        for mapping in default_mappings:
            if mapping.resource_type == self.resource_type:
                return mapping.external_id
        raise ToolkitValueError(f"No default ingestion view specified for resource type '{self.resource_type}'")

    def as_asset_centric_id(self) -> AssetCentricId:
        return AssetCentricId(resource_type=self.resource_type, id_=self.id)

    @model_validator(mode="before")
    def _handle_flat_dict(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values

        if "space" in values and "externalId" in values:
            values["instanceId"] = {"space": values.pop("space"), "externalId": values.pop("externalId")}
        if "consumerViewSpace" in values and "consumerViewExternalId" in values:
            consumer_view = {
                "space": values.pop("consumerViewSpace"),
                "externalId": values.pop("consumerViewExternalId"),
            }
            if "consumerViewVersion" in values:
                consumer_view["version"] = values.pop("consumerViewVersion")
            values["preferredConsumerView"] = consumer_view
        return values

    @field_validator("data_set_id", "ingestion_view", mode="before")
    def _empty_string_to_none(cls, v: Any) -> Any:
        if isinstance(v, str) and not v.strip():
            return None
        return v

    @field_validator("preferred_consumer_view", mode="before")
    def _validate_preferred_consumer_view(cls, v: Any) -> Any:
        if isinstance(v, dict):
            return ViewId.load(v)
        return v

    @field_validator("instance_id", mode="before")
    def _validate_instance_id(cls, v: Any) -> Any:
        if isinstance(v, dict):
            return NodeId.load(v)
        return v


class MigrationMappingList(ModelList[MigrationMapping]):
    @classmethod
    def _get_base_model_cls(cls) -> type[MigrationMapping]:
        return MigrationMapping

    @classmethod
    def _required_header_names(cls) -> set[str]:
        return {"id", "space", "externalId"}

    @classmethod
    def _optional_header_names(cls) -> set[str]:
        return {"dataSetId", "ingestionView", "consumerViewSpace", "consumerViewExternalId", "consumerViewVersion"}

    def get_ids(self) -> list[int]:
        """Return a list of IDs from the migration mappings."""
        return [mapping.id for mapping in self]

    def as_node_ids(self) -> list[NodeId]:
        """Return a list of NodeIds from the migration mappings."""
        return [mapping.instance_id for mapping in self]

    def spaces(self) -> set[str]:
        """Return a set of spaces from the migration mappings."""
        return {mapping.instance_id.space for mapping in self}

    def as_pending_ids(self) -> list[PendingInstanceId]:
        return [PendingInstanceId(pending_instance_id=mapping.instance_id, id=mapping.id) for mapping in self]

    def get_data_set_ids(self) -> set[int]:
        """Return a list of data set IDs from the migration mappings."""
        return {mapping.data_set_id for mapping in self if mapping.data_set_id is not None}

    def as_mapping_by_id(self) -> dict[int, MigrationMapping]:
        """Return a mapping of IDs to MigrationMapping objects."""
        return {mapping.id: mapping for mapping in self}

    @classmethod
    def read_csv_file(cls, filepath: Path, resource_type: AssetCentricKind | None = None) -> "MigrationMappingList":
        if cls is not MigrationMappingList or resource_type is None:
            return super().read_csv_file(filepath)
        cls_by_resource_type: dict[str, type[MigrationMappingList]] = {
            "Assets": AssetMigrationMappingList,
            "TimeSeries": TimeSeriesMigrationMappingList,
            "FileMetadata": FileMigrationMappingList,
            "Events": EventMigrationMappingList,
        }
        if resource_type not in cls_by_resource_type:
            raise ToolkitValueError(
                f"Invalid resource type '{resource_type}'. Must be one of 'asset', 'timeseries', or 'file'."
            )
        return cls_by_resource_type[resource_type].read_csv_file(filepath, resource_type=None)


class AssetMapping(MigrationMapping):
    resource_type: Literal["asset"] = "asset"


class EventMapping(MigrationMapping):
    resource_type: Literal["event"] = "event"


class TimeSeriesMapping(MigrationMapping):
    resource_type: Literal["timeseries"] = "timeseries"


class FileMapping(MigrationMapping):
    resource_type: Literal["file"] = "file"


class AssetMigrationMappingList(MigrationMappingList):
    @classmethod
    def _get_base_model_cls(cls) -> type[AssetMapping]:
        return AssetMapping


class EventMigrationMappingList(MigrationMappingList):
    @classmethod
    def _get_base_model_cls(cls) -> type[EventMapping]:
        return EventMapping


class FileMigrationMappingList(MigrationMappingList):
    @classmethod
    def _get_base_model_cls(cls) -> type[FileMapping]:
        return FileMapping


class TimeSeriesMigrationMappingList(MigrationMappingList):
    @classmethod
    def _get_base_model_cls(cls) -> type[TimeSeriesMapping]:
        return TimeSeriesMapping


@dataclass
class AssetCentricMapping(Generic[T_AssetCentricResource], WriteableCogniteResource[InstanceApply]):
    mapping: MigrationMapping
    resource: T_AssetCentricResource

    def as_write(self) -> InstanceApply:
        raise NotImplementedError()

    def dump(self, camel_case: bool = True) -> dict[str, JsonVal]:
        mapping = self.mapping.model_dump(exclude_unset=True, by_alias=camel_case)
        # Ensure that resource type is always included, even if unset.
        mapping["resourceType" if camel_case else "resource_type"] = self.mapping.resource_type
        return {
            "mapping": mapping,
            "resource": self.resource.dump(camel_case=camel_case),
        }


class AssetCentricMappingList(WriteableCogniteResourceList[InstanceApply, AssetCentricMapping[T_AssetCentricResource]]):
    _RESOURCE: type = AssetCentricMapping

    def as_write(self) -> InstanceApplyList:
        return InstanceApplyList([item.as_write() for item in self])



from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.constants import StreamTemplateName

from .base import BaseModelResource, ToolkitResource


class StreamSettings(BaseModelResource):
    """Stream settings resource class."""

    template: dict[Literal["name"], StreamTemplateName] = Field(
        description="Reference to a template which should be used to define initial settings for the stream."
    )


class StreamYAML(ToolkitResource):
    """Stream YAML resource class."""

    external_id: str = Field(
        description="The external ID of the stream.",
        min_length=1,
        max_length=100,
        pattern="^[a-z]([a-z0-9_-]{0,98}[a-z0-9])?$",
    )
    settings: StreamSettings = Field(
        description="Stream settings which should be applied to a stream.",
    )



from collections.abc import Hashable, Iterable, Sequence, Sized
from typing import Any, final

from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability
from cognite.client.data_classes.data_modeling import NodeApplyList, NodeList, ViewId
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.migration import (
    ResourceViewMapping,
    ResourceViewMappingApply,
)
from cognite_toolkit._cdf_tk.constants import COGNITE_MIGRATION_SPACE
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import ResourceViewMappingYAML
from cognite_toolkit._cdf_tk.utils import in_dict

from .datamodel import SpaceCRUD, ViewCRUD


@final
class ResourceViewMappingCRUD(
    ResourceCRUD[str, ResourceViewMappingApply, ResourceViewMapping, NodeApplyList, NodeList[ResourceViewMapping]]
):
    folder_name = "migration"
    filename_pattern = (
        r"^.*\.ResourceViewMapping$"  # Matches all yaml files whose stem ends with '.ResourceViewMapping$'.
    )
    filetypes = frozenset({"yaml", "yml"})
    resource_cls = ResourceViewMapping
    resource_write_cls = ResourceViewMappingApply
    list_cls = NodeList[ResourceViewMapping]
    list_write_cls = NodeApplyList
    kind = "ResourceViewMapping"
    dependencies = frozenset({SpaceCRUD, ViewCRUD})
    _doc_url = "Instances/operation/applyNodeAndEdges"
    yaml_cls = ResourceViewMappingYAML

    @property
    def display_name(self) -> str:
        return "resource view mapping"

    @classmethod
    def get_id(cls, item: ResourceViewMapping | ResourceViewMappingApply | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ResourceViewMappingApply] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [capabilities.DataModelInstancesAcl.Action.Read]
            if read_only
            else [capabilities.DataModelInstancesAcl.Action.Read, capabilities.DataModelInstancesAcl.Action.Write]
        )

        return capabilities.DataModelInstancesAcl(
            actions=actions, scope=capabilities.DataModelInstancesAcl.Scope.SpaceID([COGNITE_MIGRATION_SPACE])
        )

    def prerequisite_warning(self) -> str | None:
        view_id = ResourceViewMapping.get_source()
        views = self.client.data_modeling.views.retrieve(view_id)
        if len(views) > 0:
            return None
        return f"{self.display_name} requires the {view_id!r} to be deployed. run `cdf migrate prepare` to deploy it."

    def create(self, items: NodeApplyList) -> Sized:
        return self.client.migration.resource_view_mapping.upsert(items)

    def update(self, items: NodeApplyList) -> Sized:
        return self.client.migration.resource_view_mapping.upsert(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> NodeList[ResourceViewMapping]:
        return self.client.migration.resource_view_mapping.retrieve(external_id=ids)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        result = self.client.migration.resource_view_mapping.delete(external_id=ids)
        return len(result)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[ResourceViewMapping]:
        if space == COGNITE_MIGRATION_SPACE:
            return self.client.migration.resource_view_mapping.list(limit=-1)
        else:
            return []

    @classmethod
    def get_dependent_items(cls, item: dict) -> "Iterable[tuple[type[ResourceCRUD], Hashable]]":
        yield SpaceCRUD, COGNITE_MIGRATION_SPACE

        yield ViewCRUD, ResourceViewMapping.get_source()

        if "viewId" in item:
            view_id = item["viewId"]
            if isinstance(view_id, dict) and in_dict(("space", "externalId"), view_id):
                yield ViewCRUD, ViewId.load(view_id)

    def dump_resource(self, resource: ResourceViewMapping, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump(context="local")
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)
        return dumped



from abc import ABC
from pathlib import Path
from typing import Any, Literal

from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricKind

from ._base import DataSelector


class AssetCentricSelector(DataSelector, ABC):
    kind: AssetCentricKind


class DataSetSelector(AssetCentricSelector):
    """Select data associated with a specific data set."""

    type: Literal["dataSet"] = "dataSet"

    data_set_external_id: str

    @property
    def group(self) -> str:
        return f"DataSet_{self.data_set_external_id}"

    def __str__(self) -> str:
        return self.kind

    def as_filter(self) -> dict[str, Any]:
        return {"dataSetIds": [{"externalId": self.data_set_external_id}]}


class AssetSubtreeSelector(AssetCentricSelector):
    """Select data associated with an asset and its subtree."""

    type: Literal["assetSubtree"] = "assetSubtree"
    hierarchy: str

    @property
    def group(self) -> str:
        return f"Hierarchy_{self.hierarchy}"

    def __str__(self) -> str:
        return self.kind

    def as_filter(self) -> dict[str, Any]:
        return {"assetSubtreeIds": [{"externalId": self.hierarchy}]}


class AssetCentricFileSelector(AssetCentricSelector):
    """Select data from a specific file."""

    type: Literal["assetFile"] = "assetFile"
    datafile: Path

    @property
    def group(self) -> str:
        return "AssetCentric"

    def __str__(self) -> str:
        return f"file_{self.datafile.name}"



from pathlib import Path
from typing import Any

import pytest
from cognite.client.data_classes import DataSet, DataSetList
from cognite.client.data_classes.aggregations import UniqueResult, UniqueResultList
from cognite.client.data_classes.data_modeling import DataModel, NodeApply, SpaceApply, View

from cognite_toolkit._cdf_tk.commands._migrate.command import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.creators import InstanceSpaceCreator, SourceSystemCreator
from cognite_toolkit._cdf_tk.commands._migrate.data_model import COGNITE_MIGRATION_MODEL, VIEWS
from cognite_toolkit._cdf_tk.data_classes import ResourceDeployResult
from cognite_toolkit._cdf_tk.exceptions import ToolkitRequiredValueError
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestCreator:
    @pytest.mark.parametrize("dry_run", [pytest.param(True, id="dry_run"), pytest.param(False, id="not_dry_run")])
    def test_create_instance_spaces(
        self, dry_run: bool, toolkit_client_approval: ApprovalToolkitClient, tmp_path: Path
    ) -> None:
        toolkit_client_approval.append(DataModel, COGNITE_MIGRATION_MODEL)
        toolkit_client_approval.append(View, VIEWS)
        data_sets = DataSetList(
            [
                DataSet(
                    external_id=f"dataset_{letter}",
                    name=f"Dataset {letter}",
                    description=f"This is dataset {letter}",
                )
                for letter in "ABC"
            ]
        )
        toolkit_client_approval.append(DataSet, data_sets)

        results = MigrationCommand(silent=True).create(
            client=toolkit_client_approval.client,
            creator=InstanceSpaceCreator(
                toolkit_client_approval.client, data_set_external_ids=[ds.external_id for ds in data_sets]
            ),
            dry_run=dry_run,
            verbose=False,
            output_dir=tmp_path,
        )
        assert "spaces" in results
        result = results["spaces"]
        assert isinstance(result, ResourceDeployResult)
        assert result.created == 3
        configurations = list(tmp_path.rglob("*Space.yaml"))
        assert len(configurations) == 3
        created_spaces = toolkit_client_approval.created_resources["Space"] if not dry_run else []
        assert all(isinstance(space, SpaceApply) for space in created_spaces)
        expected_created = {ds.external_id for ds in data_sets} if not dry_run else set()
        assert {space.space for space in created_spaces} == expected_created

    def test_create_instance_spaces_missing_external_id(
        self, toolkit_client_approval: ApprovalToolkitClient, tmp_path: Path
    ) -> None:
        toolkit_client_approval.append(DataModel, COGNITE_MIGRATION_MODEL)
        toolkit_client_approval.append(View, VIEWS)
        data_sets = DataSetList(
            [
                DataSet(
                    id=i,
                    name=f"Dataset {i}",
                    description=f"This is dataset {i}",
                )
                for i in range(3)
            ]
        )

        with pytest.raises(
            ToolkitRequiredValueError,
            match="Cannot create instance spaces for datasets with missing external IDs: 0, 1 and 2",
        ):
            MigrationCommand(silent=True).create(
                client=toolkit_client_approval.client,
                creator=InstanceSpaceCreator(toolkit_client_approval.client, datasets=data_sets),
                dry_run=False,
                verbose=False,
                output_dir=tmp_path,
            )

    @pytest.mark.parametrize(
        "arguments",
        [
            pytest.param({"data_set_external_id": "my_data_set"}, id="with_data_set"),
            pytest.param({"hierarchy": "my_root_asset"}, id="with_hierarchy"),
        ],
    )
    def test_create_source_systems(
        self, arguments: dict[str, Any], toolkit_client_approval: ApprovalToolkitClient, tmp_path: Path
    ) -> None:
        toolkit_client_approval.append(DataModel, COGNITE_MIGRATION_MODEL)
        toolkit_client_approval.append(View, VIEWS)
        asset_sources = UniqueResultList([UniqueResult(100, ["aveva"]), UniqueResult(50, ["custom"])])
        event_sources = UniqueResultList([UniqueResult(400, ["sap"]), UniqueResult(200, ["internal"])])
        file_sources = UniqueResultList([UniqueResult(1000, ["sharepoint"])])
        client = toolkit_client_approval.mock_client
        client.assets.aggregate_unique_values.return_value = asset_sources
        client.events.aggregate_unique_values.return_value = event_sources
        client.documents.aggregate_unique_values.return_value = file_sources

        results = MigrationCommand(silent=True).create(
            client=toolkit_client_approval.client,
            creator=SourceSystemCreator(toolkit_client_approval.client, "my_source_space", **arguments),
            dry_run=False,
            verbose=False,
            output_dir=tmp_path,
        )
        assert "nodes" in results
        result = results["nodes"]
        assert isinstance(result, ResourceDeployResult)
        assert result.created == 5
        configurations = list(tmp_path.rglob("*Node.yaml"))
        assert len(configurations) == 5
        expected_external_ids = {"aveva", "custom", "sap", "internal", "sharepoint"}
        created_nodes = toolkit_client_approval.created_resources["Node"]
        assert all(isinstance(node, NodeApply) for node in created_nodes)
        assert {node.external_id for node in created_nodes} == expected_external_ids



from abc import ABC
from datetime import date, datetime, timezone

import pytest
from cognite.client.data_classes import Label, LabelDefinition
from cognite.client.data_classes.data_modeling import ContainerId
from cognite.client.data_classes.data_modeling.data_types import (
    Boolean,
    DirectRelation,
    DirectRelationReference,
    Enum,
    EnumValue,
    Float32,
    Float64,
    Int32,
    Int64,
    Json,
    PropertyType,
    Text,
    Timestamp,
)
from cognite.client.data_classes.data_modeling.instances import PropertyValueWrite

from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.dtype_conversion import (
    CONVERTER_BY_DTYPE,
    DATATYPE_CONVERTER_BY_DATA_TYPE,
    asset_centric_convert_to_primary_property,
    convert_str_to_data_type,
    convert_to_primary_property,
    infer_data_type_from_value,
)
from cognite_toolkit._cdf_tk.utils.useful_types import AVAILABLE_DATA_TYPES, AssetCentricType, DataType


class TestConvertToContainerProperty:
    @pytest.mark.parametrize(
        "value, type_, nullable, expected_value",
        [
            pytest.param(
                "string_value",
                Text(),
                True,
                "string_value",
                id="String to text",
            ),
            pytest.param(
                True,
                Boolean(),
                True,
                True,
                id="Bool to boolean",
            ),
            pytest.param(
                42,
                Int32(),
                True,
                42,
                id="Int to Int32",
            ),
            pytest.param(
                1234567890123,
                Int64(),
                True,
                1234567890123,
                id="Int to Int64",
            ),
            pytest.param(
                3.14,
                Float32(),
                True,
                3.14,
                id="Float to Float32",
            ),
            pytest.param(
                2.7182818284,
                Float64(),
                True,
                2.7182818284,
                id="Float to Float64",
            ),
            pytest.param(
                {"key": "value"},
                Json(),
                True,
                {"key": "value"},
                id="Dict to Json",
            ),
            pytest.param(
                "2025-07-22T12:34:56Z",
                Timestamp(),
                True,
                datetime(2025, 7, 22, 12, 34, 56, tzinfo=timezone.utc),
                id="String to Timestamp",
            ),
            pytest.param(
                "ENUM_A",
                Enum(values={"ENUM_A": EnumValue(), "ENUM_B": EnumValue()}),
                True,
                "ENUM_A",
                id="String to Enum",
            ),
            pytest.param(
                "1",
                Boolean(),
                True,
                True,
                id="String '1' to boolean True",
            ),
            pytest.param(
                "0",
                Boolean(),
                True,
                False,
                id="String '0' to boolean False",
            ),
            pytest.param(
                "-42",
                Int32(),
                True,
                -42,
                id="String '-42' to Int32",
            ),
            pytest.param(
                "0",
                Int64(),
                True,
                0,
                id="String '0' to Int64",
            ),
            pytest.param(
                "-3.14",
                Float32(),
                True,
                -3.14,
                id="String '-3.14' to Float32",
            ),
            pytest.param(
                "0.0",
                Float64(),
                True,
                0.0,
                id="String '0.0' to Float64",
            ),
            pytest.param(
                "[1, 2, 3]",
                Json(),
                True,
                [1, 2, 3],
                id="Stringified list to Json",
            ),
            pytest.param(
                '{"a": 1, "b": 2}',
                Json(),
                True,
                {"a": 1, "b": 2},
                id="Stringified dict with ints to Json",
            ),
            pytest.param(
                "2025-01-01T00:00:00Z",
                Timestamp(),
                True,
                datetime(2025, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                id="String to Timestamp (start of year)",
            ),
            pytest.param(
                "ENUM_B",
                Enum(values={"ENUM_A": EnumValue(), "ENUM_B": EnumValue()}),
                True,
                "ENUM_B",
                id="String to Enum (B)",
            ),
            pytest.param(
                1753193696789,
                Timestamp(),
                True,
                datetime(2025, 7, 22, 14, 14, 56, 789000, tzinfo=timezone.utc),
                id="Epoch float with milliseconds to Timestamp",
            ),
            pytest.param(
                [1, 2, 3],
                Int32(is_list=True),
                True,
                [1, 2, 3],
                id="List of int32 to Int32 list property",
            ),
            pytest.param(
                ["2025-07-22T12:34:56Z", "2025-01-01T00:00:00Z"],
                Timestamp(is_list=True),
                True,
                [
                    datetime(2025, 7, 22, 12, 34, 56, tzinfo=timezone.utc),
                    datetime(2025, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                ],
                id="List of ISO timestamps to Timestamp list property",
            ),
            # List property type valid cases (additional)
            pytest.param(
                "[1, 2, 3]",
                Int32(is_list=True),
                True,
                [1, 2, 3],
                id="JSON string list to Int32 list property",
            ),
            pytest.param(
                '["2025-07-22T12:34:56Z", "2025-01-01T00:00:00Z"]',
                Timestamp(is_list=True),
                True,
                [
                    datetime(2025, 7, 22, 12, 34, 56, tzinfo=timezone.utc),
                    datetime(2025, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                ],
                id="JSON string list to Timestamp list property",
            ),
            pytest.param(
                42,
                Int32(is_list=True),
                True,
                [42],
                id="Single int to Int32 list property",
            ),
            pytest.param(
                "2025-07-22T12:34:56Z",
                Timestamp(is_list=True),
                True,
                [datetime(2025, 7, 22, 12, 34, 56, tzinfo=timezone.utc)],
                id="Single ISO timestamp to Timestamp list property",
            ),
            pytest.param(
                1,
                DirectRelation(),
                True,
                DirectRelationReference(space="my_space", external_id="parent1"),
                id="Int to DirectRelation with cache lookup (1)",
            ),
        ],
    )
    def test_valid_conversion(
        self,
        value: str | int | float | bool | dict | list,
        type_: PropertyType,
        nullable: bool,
        expected_value: PropertyValueWrite,
    ):
        cache = {
            1: DirectRelationReference(space="my_space", external_id="parent1"),
            2: DirectRelationReference(space="my_space", external_id="parent2"),
        }
        actual = convert_to_primary_property(value, type_, nullable, direct_relation_lookup=cache)

        if isinstance(expected_value, float):
            assert actual == pytest.approx(expected_value), f"Expected {expected_value}, but got {actual}"
        else:
            assert actual == expected_value, f"Expected {expected_value}, but got {actual}"

    @pytest.mark.parametrize(
        "value, type_, nullable, error_message",
        [
            pytest.param(
                None,
                Text(),
                False,
                "Cannot convert None to a non-nullable property.",
                id="None to non-nullable Text",
            ),
            pytest.param(
                "invalid_enum_value",
                Enum(values={"ENUM_A": EnumValue(), "ENUM_B": EnumValue()}),
                True,
                "Value 'invalid_enum_value' is not a valid enum value. Available values: ENUM_A and ENUM_B",
                id="Invalid string to Enum",
            ),
            pytest.param(
                "not_a_number",
                Int32(),
                True,
                "Cannot convert not_a_number to int32.",
                id="Invalid string to Int32",
            ),
            pytest.param(
                "not_a_float",
                Float32(),
                True,
                "Cannot convert not_a_float to float32.",
                id="Invalid string to Float32",
            ),
            pytest.param(
                {"key": "value"},
                Boolean(),
                True,
                "Cannot convert {'key': 'value'} to boolean.",
                id="Dict to Boolean",
            ),
            pytest.param(
                str(2**31),
                Int32(),
                True,
                "Value 2147483648 is out of range for int32.",
                id="Int32 overflow (too large)",
            ),
            pytest.param(
                str(-(2**31) - 1),
                Int32(),
                True,
                "Value -2147483649 is out of range for int32.",
                id="Int32 underflow (too small)",
            ),
            pytest.param(
                str(3.5e38),
                Float32(),
                True,
                "Value 3.5e+38 is out of range for float32.",
                id="Float32 overflow (too large)",
            ),
            pytest.param(
                str(-3.5e38),
                Float32(),
                True,
                "Value -3.5e+38 is out of range for float32.",
                id="Float32 underflow (too small)",
            ),
            pytest.param(
                "3.14",
                Int32(),
                True,
                "Cannot convert 3.14 to int32.",
                id="Float string to Int32 (invalid)",
            ),
            pytest.param(
                "true",
                Int32(),
                True,
                "Cannot convert true to int32.",
                id="Boolean string to Int32 (invalid)",
            ),
            pytest.param(
                "ENUM_C",
                Enum(values={"ENUM_A": EnumValue(), "ENUM_B": EnumValue()}),
                True,
                "Value 'enum_c' is not a valid enum value. Available values: ENUM_A and ENUM_B",
                id="Invalid enum value (not in list)",
            ),
            pytest.param(
                '{"key": "value"}',
                Int32(),
                True,
                'Cannot convert {"key": "value"} to int32.',
                id="JSON dict string to Int32 (invalid)",
            ),
            pytest.param(
                "1e100",
                Float32(),
                True,
                "Value 1e100 is out of range for float32.",
                id="Scientific notation overflow for Float32",
            ),
            pytest.param(
                "9223372036854775808",  # 2**63
                Int64(),
                True,
                "Value 9223372036854775808 is out of range for int64.",
                id="Int64 overflow (too large)",
            ),
            pytest.param(
                "-9223372036854775809",  # -(2**63) - 1
                Int64(),
                True,
                "Value -9223372036854775809 is out of range for int64.",
                id="Int64 underflow (too small)",
            ),
            pytest.param(
                [123, 456],
                Int64(is_list=False),
                True,
                "Expected a single value for int64, but got a list.",
                id="List to Int64 (invalid, not a list type)",
            ),
            pytest.param(
                1,
                DirectRelation(),
                True,
                "Cannot convert 1 to DirectRelationReference. Invalid data type or missing in lookup.",
                id="DirectRelation with cache miss",
            ),
            pytest.param(
                True,
                DirectRelation(),
                True,
                "Cannot convert True to DirectRelationReference. Invalid data type or missing in lookup.",
                id="DirectRelation with invalid type (bool instead of str or int)",
            ),
        ],
    )
    def test_invalid_conversion(
        self, value: str | int | float | bool | dict | list, type_: PropertyType, nullable: bool, error_message: str
    ):
        with pytest.raises(ValueError) as exc_info:
            convert_to_primary_property(value, type_, nullable, direct_relation_lookup={})

        assert str(exc_info.value) == error_message, (
            f"Expected error message '{error_message}', but got '{exc_info.value}'"
        )

    def test_all_converters_registered(self) -> None:
        """Checks that all property types that are in the cognite-sdk have a corresponding converter."""
        existing_types: set[str] = set()
        to_check = [PropertyType]
        while to_check:
            current_type = to_check.pop()
            for subclass in current_type.__subclasses__():
                if hasattr(subclass, "_type"):
                    existing_types.add(subclass._type)
                if ABC in subclass.__bases__:
                    to_check.append(subclass)

        missing_converters = existing_types - set(CONVERTER_BY_DTYPE.keys())

        assert not missing_converters, (
            f"Missing converters for types: {humanize_collection(missing_converters)}. "
            "Please ensure all property types have a corresponding converter."
        )

    @pytest.mark.parametrize(
        "value, type_, destination_container_property, source_property, expected",
        [
            pytest.param(
                True,
                Enum(values={"numeric": EnumValue(), "string": EnumValue()}),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "type"),
                ("timeseries", "isString"),
                "string",
                id="TimeSeries.isString to Enum conversion",
            ),
            pytest.param(
                False,
                Enum(values={"numeric": EnumValue(), "string": EnumValue()}),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "type"),
                ("timeseries", "isString"),
                "numeric",
                id="TimeSeries.isString to Enum conversion (False case)",
            ),
            pytest.param(
                [Label("pump"), Label("mechanical")],
                Text(is_list=True),
                (ContainerId("cdf_cdm", "CogniteDescribable"), "tags"),
                ("asset", "labels"),
                ["pump", "mechanical"],
                id="Asset labels to tags list conversion",
            ),
            pytest.param(
                [Label("pump"), {"externalId": "mechanical"}, LabelDefinition("equipment")],
                Text(is_list=True),
                (ContainerId("cdf_cdm", "CogniteDescribable"), "tags"),
                ("file", "labels"),
                ["pump", "mechanical", "equipment"],
                id="Asset label to tags list conversion",
            ),
            pytest.param(
                False,
                Boolean(),
                (ContainerId("some_other_space", "SomeOtherView"), "type"),
                ("timeseries", "isString"),
                False,
                id="Non-asset-centric boolean to primary property conversion",
            ),
            pytest.param(
                "acceleration:m-per-sec2",
                DirectRelation(),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "unit"),
                ("timeseries", "unitExternalId"),
                {"space": "cdf_cdm_units", "externalId": "acceleration:m-per-sec2"},
                id="TimeSeries unitExternalId to DirectRelation conversion",
            ),
            pytest.param(
                None,
                DirectRelation(),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "unit"),
                ("timeseries", "unitExternalId"),
                None,
                id="TimeSeries unitExternalId to DirectRelation conversion with None value (nullable)",
            ),
            pytest.param(
                "source1",
                DirectRelation(),
                (ContainerId("cdf_cdm", "CogniteSourceable"), "source"),
                ("timeseries", "source"),
                DirectRelationReference(space="spaceA", external_id="source1"),
                id="TimeSeries sourceExternalId to DirectRelation conversion with cache lookup",
            ),
        ],
    )
    def test_asset_centric_conversion(
        self,
        value: str | int | float | bool | dict | list,
        type_: PropertyType,
        destination_container_property: tuple[ContainerId, str],
        source_property: tuple[AssetCentricType, str],
        expected: PropertyValueWrite,
    ):
        cache = {
            "source1": DirectRelationReference(space="spaceA", external_id="source1"),
            "source2": DirectRelationReference(space="spaceB", external_id="source2"),
        }

        actual = asset_centric_convert_to_primary_property(
            value, type_, True, destination_container_property, source_property, direct_relation_lookup=cache
        )

        assert actual == expected

    @pytest.mark.parametrize(
        "value, type_, destination_container_property, source_property, error_message",
        [
            pytest.param(
                "invalid_value",
                Enum(values={"numeric": EnumValue(), "string": EnumValue()}),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "type"),
                ("timeseries", "isString"),
                "Cannot convert invalid_value to TimeSeries type. Expected a boolean value.",
                id="Invalid TimeSeries type input value conversion error",
            ),
            pytest.param(
                "not_a_list",
                Text(is_list=True),
                (ContainerId("cdf_cdm", "CogniteDescribable"), "tags"),
                ("asset", "labels"),
                "Cannot convert not_a_list to labels. Expected a list of Labels, objects, or LabelDefinitions.",
                id="List to Text conversion error",
            ),
            pytest.param(
                True,
                DirectRelation(),
                (ContainerId("cdf_cdm", "CogniteTimeSeries"), "unit"),
                ("timeseries", "unitExternalId"),
                "Cannot convert True to TimeSeries unit. Expected a string representing the externalId.",
                id="TimeSeries unitExternalId to DirectRelation conversion error",
            ),
            pytest.param(
                "unknown_source",
                DirectRelation(),
                (ContainerId("cdf_cdm", "CogniteSourceable"), "source"),
                ("timeseries", "source"),
                "Cannot convert 'unknown_source' to DirectRelationReference. Invalid data type or missing in lookup.",
                id="TimeSeries sourceExternalId to DirectRelation conversion with missing cache entry",
            ),
        ],
    )
    def test_asset_centric_failed_conversion(
        self,
        value: str | int | float | bool | dict | list,
        type_: PropertyType,
        destination_container_property: tuple[ContainerId, str],
        source_property: tuple[AssetCentricType, str],
        error_message: str,
    ):
        with pytest.raises(ValueError) as exc_info:
            asset_centric_convert_to_primary_property(
                value, type_, True, destination_container_property, source_property, direct_relation_lookup={}
            )

        assert str(exc_info.value) == error_message


class TestConvertStringToDataType:
    TEST_CASES = (
        pytest.param("string_value", "string", "string_value", id="String to Text"),
        pytest.param("42", "integer", 42, id="String to integer"),
        pytest.param("3.14", "float", 3.14, id="String to float"),
        pytest.param("true", "boolean", True, id="String 'true' to Boolean"),
        pytest.param("false", "boolean", False, id="String 'false' to Boolean"),
        pytest.param('{"key": "value"}', "json", {"key": "value"}, id="Stringified dict to Json"),
        pytest.param(
            "2025-07-22T12:34:56Z",
            "timestamp",
            datetime(2025, 7, 22, 12, 34, 56, tzinfo=timezone.utc),
            id="ISO timestamp to Timestamp",
        ),
        pytest.param("2025-07-22", "date", date(2025, 7, 22), id="ISO date to Date"),
    )

    @pytest.mark.parametrize(
        "value, data_type, expected_value",
        (*TEST_CASES, pytest.param(None, "string", None, id="None to string (nullable)")),
    )
    def test_convert(
        self,
        value: str | None,
        data_type: DataType,
        expected_value: str | int | float | bool | dict | list | datetime | date | None,
    ) -> None:
        result = convert_str_to_data_type(value, data_type)
        if isinstance(expected_value, float):
            assert result == pytest.approx(expected_value), f"Expected {expected_value}, but got {result}"
        else:
            assert result == expected_value, f"Expected {expected_value}, but got {result}"

    @pytest.mark.parametrize(
        "value, data_type, expected_value",
        (*TEST_CASES, pytest.param(None, "string", [], id="None to string (nullable)")),
    )
    def test_convert_array(
        self,
        value: str | None,
        data_type: DataType,
        expected_value: str | int | float | bool | dict | list | datetime | date | None,
    ) -> None:
        result = convert_str_to_data_type(value, data_type, is_array=True)
        if isinstance(expected_value, float):
            assert result == pytest.approx([expected_value]), f"Expected [{expected_value}], but got {result}"
        elif isinstance(expected_value, list):
            assert result == expected_value, f"Expected {expected_value}, but got {result}"
        else:
            assert result == [expected_value], f"Expected [{expected_value}], but got {result}"

    def test_all_data_type_converters_registered(self) -> None:
        """Checks that all data types that are in the toolkit have a corresponding converter."""
        assert AVAILABLE_DATA_TYPES == set(DATATYPE_CONVERTER_BY_DATA_TYPE.keys())


class TestInferDataTypeFromValue:
    TEST_CASES = (
        pytest.param("string_value", "string", id="String to string"),
        pytest.param("42", "integer", id="String to integer"),
        pytest.param("3.14", "float", id="String to float"),
        pytest.param("true", "boolean", id="String 'true' to boolean"),
        pytest.param("false", "boolean", id="String 'false' to boolean"),
        pytest.param('{"key": "value"}', "json", id="Stringified dict to json"),
        pytest.param("2025-07-22T12:34:56Z", "timestamp", id="ISO timestamp to timestamp"),
        pytest.param("2025-07-22", "date", id="ISO date to date"),
        # Numeric Edge Cases
        pytest.param("42.0", "float", id="Float with zero decimal"),
        pytest.param("-10", "integer", id="Negative integer"),
        # Date/Time Ambiguity
        pytest.param("2025-07-22T00:00:00Z", "timestamp", id="Midnight UTC timestamp should be timestamp"),
        # JSON Variations
        pytest.param("[]", "json", id="JSON array"),
        # Malformed Inputs (Fallback to string)
        pytest.param("3.14.15.16.17", "string", id="Malformed float falls back to string"),
        pytest.param('{"key": "value', "string", id="Malformed JSON falls back to string"),
        pytest.param("2025-99-99", "string", id="Malformed date falls back to string"),
        pytest.param("", "string", id="Empty string is string"),
    )

    @pytest.mark.parametrize("value, expected_type", TEST_CASES)
    def test_infer_data_type_from_value(self, value: str, expected_type: str) -> None:
        result, _ = infer_data_type_from_value(value, dtype="Python")
        assert result == expected_type, f"Expected {expected_type}, but got {result}"

    @pytest.mark.parametrize("value, expected_type", TEST_CASES)
    def test_infer_data_type_from_value_json(self, value: str, expected_type: str) -> None:
        expected_type = {"timestamp": "string", "date": "string"}.get(expected_type, expected_type)
        result, _ = infer_data_type_from_value(value, dtype="Json")
        assert result == expected_type, f"Expected {expected_type}, but got {result}"



from pydantic import Field

from cognite_toolkit._cdf_tk.resource_classes.base import BaseModelResource, ToolkitResource


class Translation(BaseModelResource):
    x: float = Field(description="X coordinate")
    y: float = Field(description="Y coordinate")
    z: float = Field(description="Z coordinate")


class Orientation(Translation):
    w: float = Field(description="W coordinate")


class Transform(BaseModelResource):
    parent_frame_external_id: str = Field(description="Parent frame external id.")
    translation: Translation = Field(description="Translation")
    orientation: Orientation = Field(description="Orientation")


class RobotFrameYAML(ToolkitResource):
    name: str = Field(description="Robot frame name.")
    external_id: str = Field(description="RobotFrame external id.", max_length=256)
    transform: Transform | None = Field(default=None, description="Transform")



import sys
from collections.abc import Collection, Iterator
from typing import Any, Generic, Protocol, TypeVar

from cognite.client import CogniteClient

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class ResourceRequestProtocol(Protocol):
    @classmethod
    def _load(cls, data: dict[str, Any]) -> Self: ...

    def dump(self, camel_case: bool = True) -> dict[str, Any]: ...


class ResourceResponseProtocol(Protocol):
    def as_write(self) -> ResourceRequestProtocol: ...


T_ResourceRequest = TypeVar("T_ResourceRequest", bound=ResourceRequestProtocol)
T_ResourceResponse = TypeVar("T_ResourceResponse", bound=ResourceResponseProtocol)


class ResourceRequestListProtocol(Protocol, Generic[T_ResourceRequest]):
    def __init__(self, collection: Collection[T_ResourceRequest]): ...

    def __getitem__(self, index: int) -> T_ResourceRequest: ...

    def __setitem__(self, index: int, value: T_ResourceRequest) -> None: ...

    def __delitem__(self, index: int) -> None: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[T_ResourceRequest]: ...

    def __contains__(self, value: object) -> bool: ...

    def __reversed__(self) -> Iterator[T_ResourceRequest]: ...

    def insert(self, index: int, value: T_ResourceRequest) -> None: ...

    def append(self, value: T_ResourceRequest) -> None: ...

    def extend(self, values: Collection[T_ResourceRequest]) -> None: ...

    def pop(self, index: int = -1) -> T_ResourceRequest: ...

    def remove(self, value: T_ResourceRequest) -> None: ...

    def clear(self) -> None: ...

    @classmethod
    def load(cls, data: list[dict[str, Any]], cognite_client: CogniteClient | None = None) -> Self: ...


class ResourceResponseListProtocol(Protocol, Generic[T_ResourceResponse]):
    def __init__(self, collection: Collection[T_ResourceResponse]): ...

    def __getitem__(self, index: int) -> T_ResourceResponse: ...

    def __setitem__(self, index: int, value: T_ResourceResponse) -> None: ...

    def __delitem__(self, index: int) -> None: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[T_ResourceResponse]: ...

    def __contains__(self, value: object) -> bool: ...

    def __reversed__(self) -> Iterator[T_ResourceResponse]: ...

    def insert(self, index: int, value: T_ResourceResponse) -> None: ...

    def append(self, value: T_ResourceResponse) -> None: ...

    def extend(self, values: Collection[T_ResourceResponse]) -> None: ...

    def pop(self, index: int = -1) -> T_ResourceResponse: ...

    def remove(self, value: T_ResourceResponse) -> None: ...

    def clear(self) -> None: ...

    def as_write(self) -> ResourceRequestListProtocol: ...

    @classmethod
    def load(cls, data: list[dict[str, Any]], cognite_client: CogniteClient | None = None) -> Self: ...


T_ResourceRequestList = TypeVar("T_ResourceRequestList", bound=ResourceRequestListProtocol)
T_ResourceResponseList = TypeVar("T_ResourceResponseList", bound=ResourceResponseListProtocol)



from typing import Literal

from pydantic import Field

from .base import BaseModelResource, ToolkitResource


class Scenes(BaseModelResource):
    external_id: str = Field(description="External ID of the scene.")
    space: str = Field(description="The space that the scene is in")


class DataModelID(BaseModelResource):
    external_id: str = Field(description="The external ID of the data model.")
    space: str = Field(description="The space of the data model.")
    version: str = Field(description="The version of the data model.")


class LocationFilterViewId(BaseModelResource):
    external_id: str = Field(description="The external ID of the view.")
    space: str = Field(description="The space of the view.")
    version: str = Field(description="The version of the view.")
    represents_entity: Literal["MAINTENANCE_ORDER", "OPERATION", "NOTIFICATION", "ASSET"] = Field(
        description="Represents entity type"
    )


class AssetCentricFields(BaseModelResource):
    data_set_external_ids: list[str] | None = Field(default=None, description="The list of data set external IDs")
    asset_subtree_external_ids: list[dict[Literal["externalId"], str]] | None = Field(
        default=None, description="External IDs of the asset."
    )
    external_id_prefix: str | None = Field(default=None, description="The external ID prefix")


class AssetCentricResourceData(AssetCentricFields):
    """ """


class AssetCentricResource(AssetCentricFields):
    assets: AssetCentricResourceData | None = Field(default=None, description="Asset resource type data")
    events: AssetCentricResourceData | None = Field(default=None, description="Event resource type data")
    timeseries: AssetCentricResourceData | None = Field(default=None, description="Timeseries resource type data")
    files: AssetCentricResourceData | None = Field(default=None, description="File resource type data")
    sequences: AssetCentricResourceData | None = Field(default=None, description="Sequence resource type data")


class LocationYAML(ToolkitResource):
    external_id: str = Field(description="The external ID provided by the client.")
    name: str = Field(description="The name of the location.")
    description: str | None = Field(default=None, description="The description of the data set.", max_length=255)
    parent_external_id: str | None = Field(default=None, description="The external ID of the parent location.")
    data_models: list[DataModelID] | None = Field(
        default=None, description="The data models associated with the location."
    )
    instance_spaces: list[str] | None = Field(default=None, description="The list of spaces that instances are in")
    scene: Scenes | None = Field(default=None, description="The scene config for the location.")
    asset_centric: AssetCentricResource | None = Field(
        default=None,
        description="The filter definition for asset centric resource types",
    )
    views: list[LocationFilterViewId] | None = Field(
        default=None, description="The views associated with the location."
    )
    data_modeling_type: Literal["HYBRID", "DATA_MODELING_ONLY"] | None = Field(
        default="HYBRID", description="The data modeling type"
    )



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from cognite_toolkit._cdf_tk.cruds import (
    ResourceCRUD,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    WarningList,
)
from cognite_toolkit._cdf_tk.tk_warnings.fileread import (
    FileReadWarning,
)
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump

from ._built_resources import SourceLocation


@dataclass
class BuildSourceFile:
    source: SourceLocation
    content: str
    loaded: list[dict[str, Any]] | dict[str, Any] | None = None


@dataclass
class BuildDestinationFile:
    path: Path
    loaded: list[dict[str, Any]] | dict[str, Any]
    loader: type[ResourceCRUD]
    source: SourceLocation
    extra_sources: list[SourceLocation] | None
    warnings: WarningList[FileReadWarning] = field(default_factory=WarningList[FileReadWarning])

    @property
    def content(self) -> str:
        return yaml_safe_dump(self.loaded)



from unittest.mock import MagicMock

import pytest
from rich.console import Console

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils.text import (
    sanitize_instance_external_id,
    sanitize_spreadsheet_title,
    suffix_description,
)


class TestSuffixDescription:
    @pytest.mark.parametrize(
        "description, expected_description, expect_warning",
        [
            (None, "<suffix>", False),
            ("", "<suffix>", False),
            ("Existing description", "Existing description <suffix>", False),
            ("Existing description <suffix>", "Existing description <suffix>", False),
            (
                "A very long description that exceeds the limit with more than one character",
                "A very long descrip...<suffix>",
                True,
            ),
        ],
    )
    def test_suffix_description(self, description: str | None, expected_description: str, expect_warning: True) -> None:
        console = MagicMock(spec=Console)
        limit = 30
        actual = suffix_description("<suffix>", description, limit, "identifier", "my_resource_type", console=console)

        assert actual == expected_description
        assert len(actual) <= limit
        if expect_warning:
            assert console.print.call_count == 1
            message = "".join(console.print.call_args[0])
            assert "Description is too long for my_resource_type 'identifier'." in message


class TestSanitizeSpreadsheetTitle:
    @pytest.mark.parametrize(
        "title, expected",
        [
            ("Valid Title", "Valid Title"),
            ("Invalid/Title", "InvalidTitle"),
            ("Another*Invalid:Title", "AnotherInvalidTitle"),
            ("", "Sheet"),
            (None, "Sheet"),
        ],
    )
    def test_sanitize_spreadsheet_title(self, title: str | None, expected: str) -> None:
        assert sanitize_spreadsheet_title(title) == expected


class TestSanitizeInstanceExternalId:
    @pytest.mark.parametrize(
        "external_id, expected",
        [
            pytest.param("valid_external_id", "valid_external_id", id="valid_external_id"),
            pytest.param("x" * 257, "x" * 247 + "_15eb95a4", id="too_long"),
        ],
    )
    def test_sanitize_instance_external_id(self, external_id: str, expected: str) -> None:
        assert sanitize_instance_external_id(external_id) == expected

    @pytest.mark.parametrize(
        "external_id",
        [
            pytest.param("", id="empty_string"),
            pytest.param("\x00", id="null_character"),
        ],
    )
    def test_sanitize_instance_external_id_raise(self, external_id: str) -> None:
        with pytest.raises(ToolkitValueError):
            sanitize_instance_external_id(external_id)



import itertools
from datetime import datetime
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes._base import CogniteResource
from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.client.api.canvas import CanvasAPI
from cognite_toolkit._cdf_tk.client.api.extended_data_modeling import ExtendedInstancesAPI
from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    CANVAS_INSTANCE_SPACE,
    Canvas,
    CanvasAnnotation,
    CanvasAnnotationApply,
    CanvasApply,
    CogniteSolutionTag,
    CogniteSolutionTagApply,
    ContainerReference,
    ContainerReferenceApply,
    FdmInstanceContainerReference,
    FdmInstanceContainerReferenceApply,
    IndustrialCanvas,
    IndustrialCanvasApply,
)
from tests.test_unit.utils import FakeCogniteResourceGenerator


class TestLoadDump:
    @pytest.mark.parametrize(
        "node_cls",
        [
            Canvas,
            CanvasApply,
            CanvasAnnotation,
            CanvasAnnotationApply,
            CogniteSolutionTag,
            CogniteSolutionTagApply,
            FdmInstanceContainerReference,
            FdmInstanceContainerReferenceApply,
            ContainerReference,
            ContainerReferenceApply,
        ],
    )
    def test_dump_reload(self, node_cls: type[CogniteResource]) -> None:
        instance = FakeCogniteResourceGenerator().create_instance(node_cls)
        dumped = instance.dump()
        reloaded = node_cls.load(dumped)

        assert reloaded == instance, f"Expected: {instance}, but got: {reloaded}"


@pytest.fixture()
def instance_api() -> MagicMock:
    return MagicMock(spec=ExtendedInstancesAPI)


class TestCanvasAPI:
    def test_upsert(self, instance_api: MagicMock) -> None:
        canvas_api = CanvasAPI(instance_api=instance_api)

        canvas_apply = CanvasApply(
            external_id="test_canvas", name="Test Canvas", created_by="me", updated_by="me", updated_at=datetime.now()
        )
        canvas_api.upsert(canvas_apply)

        assert instance_api.apply.called
        assert instance_api.apply.call_args[0][0] == canvas_apply

    @pytest.mark.parametrize(
        "external_id, expected_arg",
        [
            ("test_canvas", [NodeId(CANVAS_INSTANCE_SPACE, "test_canvas")]),
            (
                ["canvas1", "canvas2"],
                [NodeId(CANVAS_INSTANCE_SPACE, "canvas1"), NodeId(CANVAS_INSTANCE_SPACE, "canvas2")],
            ),
        ],
    )
    def test_delete(
        self, external_id: str | list[str], expected_arg: NodeId | list[NodeId], instance_api: MagicMock
    ) -> None:
        canvas_api = CanvasAPI(instance_api=instance_api)

        canvas_api.delete(external_id)

        assert instance_api.delete.called
        assert instance_api.delete.call_args[0][0] == expected_arg

    @pytest.mark.parametrize(
        "external_id, expected_arg",
        [
            ("test_canvas", NodeId(CANVAS_INSTANCE_SPACE, "test_canvas")),
            (
                ["canvas1", "canvas2"],
                [NodeId(CANVAS_INSTANCE_SPACE, "canvas1"), NodeId(CANVAS_INSTANCE_SPACE, "canvas2")],
            ),
        ],
    )
    def test_retrieve(self, external_id: str | list[str], expected_arg: NodeId, instance_api: MagicMock) -> None:
        canvas_api = CanvasAPI(instance_api=instance_api)

        canvas_api.retrieve(external_id)

        assert instance_api.retrieve_nodes.called
        assert instance_api.retrieve_nodes.call_args[0][0] == expected_arg

    def test_list(self, instance_api: MagicMock) -> None:
        canvas_api = CanvasAPI(instance_api=instance_api)

        canvas_api.list(limit=100)

        assert instance_api.list.called
        assert instance_api.list.call_args[1] == {
            "instance_type": Canvas,
            "space": CANVAS_INSTANCE_SPACE,
            "limit": 100,
            "sort": None,
            "filter": None,
        }


class TestIndustrialCanvasDataClass:
    def test_create_backup(self) -> None:
        instance = FakeCogniteResourceGenerator().create_instance(IndustrialCanvasApply)

        backup = instance.create_backup()

        original_ids = set(instance.as_instance_ids(include_solution_tags=False))
        backup_ids = set(backup.as_instance_ids(include_solution_tags=False))
        assert len(original_ids) == len(backup_ids)
        overlapping_ids = original_ids.intersection(backup_ids)
        assert not overlapping_ids

        original_ids = set(instance.as_instance_ids(include_solution_tags=True))
        backup_ids = set(backup.as_instance_ids(include_solution_tags=True))
        overlapping_ids = original_ids.intersection(backup_ids)
        solution_tags_ids = {tag.as_id() for tag in instance.solution_tags}
        assert solution_tags_ids == overlapping_ids, "Expected overlapping IDs to be solution tags IDs only"

        existing_version = [
            item.as_id()
            for item in itertools.chain(
                [backup.canvas],
                backup.annotations,
                backup.container_references,
                backup.fdm_instance_container_references,
                # Solutions tags are expected to have existing versions as these are
                # not created, but reused in the backup
            )
            if item.existing_version is not None
        ]
        assert not existing_version, (
            f"Expected no existing versions in backup annotations. Found: {len(existing_version)}"
        )

    def test_dump_load(self) -> None:
        instance = FakeCogniteResourceGenerator().create_instance(IndustrialCanvas)
        dumped = instance.dump()
        reloaded = IndustrialCanvas.load(dumped)

        assert reloaded.dump() == instance.dump(), "Failed to reload IndustrialCanvas."

    def test_as_instances(self) -> None:
        instance = FakeCogniteResourceGenerator().create_instance(IndustrialCanvasApply)

        node_or_edges = instance.as_instances()

        # 1 canvas + solution tags + 2 x (rest - one for node and one for edge connecting canvas to node)
        expected_instance_count = (
            1
            + len(instance.solution_tags)
            + 2
            * (
                len(instance.container_references)
                + len(instance.fdm_instance_container_references)
                + len(instance.annotations)
            )
        )
        assert len(node_or_edges) == expected_instance_count

        assert len(instance.solution_tags) > 0, "Expected solution tags to be present in the instance."
        assert len(instance.as_instance_ids(include_solution_tags=False)) == expected_instance_count - len(
            instance.solution_tags
        )
        assert len(instance.as_instance_ids(include_solution_tags=True)) == expected_instance_count
        assert instance.as_id() == instance.canvas.external_id

    def test_as_write(self) -> None:
        instance = FakeCogniteResourceGenerator().create_instance(IndustrialCanvas)
        write_instances = instance.as_write()

        assert isinstance(write_instances, IndustrialCanvasApply)

        assert write_instances.canvas.external_id == instance.canvas.external_id
        assert len(write_instances.container_references) == len(instance.container_references) > 0
        assert (
            len(write_instances.fdm_instance_container_references)
            == len(instance.fdm_instance_container_references)
            > 0
        )
        assert len(write_instances.annotations) == len(instance.annotations) > 0
        assert len(write_instances.solution_tags) == len(instance.solution_tags) > 0



import pytest
from cognite.client.data_classes.data_modeling import DataModel, DataModelApply, Space, ViewId
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    LocationFilter,
    LocationFilterList,
    LocationFilterWrite,
)
from tests.test_integration.constants import RUN_UNIQUE_ID

SESSION_EXTERNAL_ID = f"loc_ext_id_{RUN_UNIQUE_ID}"


@pytest.fixture(scope="session", autouse=True)
def cleanup_before_session(toolkit_client: ToolkitClient) -> None:
    # Cleanup previously created locationfilters

    try:
        for location_filter in toolkit_client.search.locations.list():
            if location_filter.external_id.startswith("loc_ext_id_"):
                toolkit_client.search.locations.delete(location_filter.id)
    except CogniteAPIError:
        pass

    yield  #


def _get_location_filter(toolkit_client: ToolkitClient, new_loc_filter: LocationFilterWrite) -> LocationFilter:
    try:
        for loc_filter in toolkit_client.search.locations.list():
            if loc_filter.external_id == new_loc_filter.external_id:
                return loc_filter
    except CogniteAPIError:
        pass
    created = toolkit_client.search.locations.create(new_loc_filter)
    return created


@pytest.fixture(scope="session")
def existing_location_filter(toolkit_client: ToolkitClient) -> LocationFilter:
    location_filter = LocationFilterWrite(
        name="loc",
        external_id=SESSION_EXTERNAL_ID,
        data_modeling_type="DATA_MODELING_ONLY",
    )
    return _get_location_filter(toolkit_client, location_filter)


@pytest.fixture(scope="session")
def my_data_model(toolkit_client: ToolkitClient, toolkit_space: Space) -> DataModel:
    data_model = toolkit_client.data_modeling.data_models.apply(
        DataModelApply(
            space=toolkit_space.space,
            external_id="LocationFilterTest",
            version="v1",
            views=[ViewId("cdf_cdm", "CogniteTimeSeries", "v1")],
        )
    )
    return data_model


@pytest.fixture(scope="session")
def second_location_filter(toolkit_client: ToolkitClient) -> LocationFilter:
    location_filter = LocationFilterWrite(
        external_id=f"{SESSION_EXTERNAL_ID}_2",
        name="loc2",
    )
    return _get_location_filter(toolkit_client, location_filter)


class TestLocationFilterAPI:
    def test_create_retrieve_delete(self, toolkit_client: ToolkitClient, my_data_model: DataModel) -> None:
        location_filter = LocationFilterWrite(
            name="loc",
            external_id=SESSION_EXTERNAL_ID,
            data_models=[my_data_model.as_id()],
            data_modeling_type="DATA_MODELING_ONLY",
        )
        created: LocationFilter | None = None
        try:
            created = toolkit_client.search.locations.create(location_filter)
            assert isinstance(created, LocationFilter)
            assert created.as_write().dump() == location_filter.dump()
            assert created.data_modeling_type is not None

            retrieved = toolkit_client.search.locations.retrieve(created.id)
            assert isinstance(retrieved, LocationFilter)
            assert retrieved.as_write().dump() == location_filter.dump()

            toolkit_client.search.locations.delete(created.id)

            with pytest.raises(CogniteAPIError):
                toolkit_client.search.locations.retrieve(created.id)
        finally:
            if created:
                try:
                    toolkit_client.search.locations.delete(created.id)
                except CogniteAPIError:
                    pass

    def test_list_location_filters(
        self, toolkit_client: ToolkitClient, existing_location_filter: LocationFilter
    ) -> None:
        location_filters = toolkit_client.search.locations.list()
        assert isinstance(location_filters, LocationFilterList)
        assert len(location_filters) > 0

    def test_iterate_location_filters(
        self, toolkit_client: ToolkitClient, existing_location_filter: LocationFilter
    ) -> None:
        for location_filters in toolkit_client.search.locations:
            assert isinstance(location_filters, LocationFilter)
            break
        else:
            pytest.fail("No location filters found")

    def test_update_location_filter(
        self, toolkit_client: ToolkitClient, existing_location_filter: LocationFilter
    ) -> None:
        update = existing_location_filter
        update.description = "New description"
        updated = toolkit_client.search.locations.update(update.id, update.as_write())
        assert updated.description == update.description

    def test_retrieve_multiple_location_filters(
        self,
        toolkit_client: ToolkitClient,
        existing_location_filter: LocationFilter,
        second_location_filter: LocationFilter,
    ) -> None:
        ids = [existing_location_filter.id, second_location_filter.id]
        retrieved = toolkit_client.search.locations.retrieve_multiple(ids)
        assert isinstance(retrieved, LocationFilterList)
        assert len(retrieved) == 2
        assert retrieved[0].id in ids
        assert retrieved[1].id in ids

    def test_retrieve_multiple_empty_list(self, toolkit_client: ToolkitClient) -> None:
        retrieved = toolkit_client.search.locations.retrieve_multiple([])
        assert isinstance(retrieved, LocationFilterList)
        assert len(retrieved) == 0

    def test_retrieve_multiple_unknown_ids(self, toolkit_client: ToolkitClient) -> None:
        with pytest.raises(CogniteAPIError):
            toolkit_client.search.locations.retrieve_multiple([-1, -2])



import shutil
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

from cognite_toolkit._cdf_tk.builders import Builder
from cognite_toolkit._cdf_tk.constants import INDEX_PATTERN
from cognite_toolkit._cdf_tk.cruds import GraphQLCRUD
from cognite_toolkit._cdf_tk.data_classes import (
    BuildDestinationFile,
    BuildSourceFile,
    ModuleLocation,
    SourceLocation,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning


class DataModelBuilder(Builder):
    _resource_folder = GraphQLCRUD.folder_name

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | list[ToolkitWarning]]:
        graphql_files = {
            source_file.source.path: source_file
            for source_file in source_files
            if source_file.source.path.suffix == ".graphql"
        }

        for source_file in source_files:
            loaded = source_file.loaded
            if loaded is None:
                # Not a YAML file
                continue
            loader, warning = self._get_loader(source_file.source.path)
            if loader is None:
                if warning is not None:
                    yield [warning]
                continue

            destination_path = self._create_destination_path(source_file.source.path, loader.kind)

            extra_sources: list[SourceLocation] | None = None
            if loader is GraphQLCRUD:
                # The GraphQL must be copied over instead of added to the DML field as
                # it is hashed in the deployment step and used to determine if the DML has changed.
                extra_sources = self._copy_graphql_to_build(source_file, destination_path, graphql_files)

            destination = BuildDestinationFile(
                path=destination_path,
                loaded=loaded,
                loader=loader,
                source=source_file.source,
                extra_sources=extra_sources,
            )
            yield destination

    def _copy_graphql_to_build(
        self,
        source_file: BuildSourceFile,
        destination_path: Path,
        graphql_files: dict[Path, BuildSourceFile],
    ) -> list[SourceLocation]:
        extra_sources: list[SourceLocation] = []
        loaded_list: list[dict[str, Any]] = (
            source_file.loaded if isinstance(source_file.loaded, list) else [source_file.loaded]  # type: ignore[list-item]
        )

        for entry in loaded_list:
            if "dml" in entry:
                expected_filename = entry["dml"]
            else:
                expected_filename = f"{INDEX_PATTERN.sub('', source_file.source.path.stem.removesuffix(GraphQLCRUD.kind).removesuffix('.'))}.graphql"
            expected_path = source_file.source.path.parent / Path(expected_filename)

            if expected_path in graphql_files:
                shutil.copy(graphql_files[expected_path].source.path, destination_path.with_suffix(".graphql"))
                extra_sources.append(graphql_files[expected_path].source)
            else:
                raise ToolkitFileNotFoundError(
                    f"Failed to find GraphQL file. Expected {expected_filename} adjacent to {source_file.source.path.as_posix()}"
                )
        return extra_sources

    def load_extra_field(self, extra: str) -> tuple[str, Any]:
        return "dml", extra



from datetime import datetime
from pathlib import Path

import pytest
from cognite.client.data_classes import DataSet, TimeSeries, TimeSeriesWrite

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import UploadCommand
from cognite_toolkit._cdf_tk.storageio import DatapointsIO
from cognite_toolkit._cdf_tk.storageio.selectors import DataPointsFileSelector
from cognite_toolkit._cdf_tk.storageio.selectors._datapoints import ExternalIdColumn, InternalIdColumn


@pytest.fixture(scope="session")
def two_timeseries(toolkit_client: ToolkitClient, toolkit_dataset: DataSet) -> tuple[TimeSeries, TimeSeries]:
    ts = TimeSeriesWrite(
        name="Test TimeSeries for Datapoints IO",
        external_id="test_timeseries_datapoints_io",
        is_step=False,
        is_string=False,
        data_set_id=toolkit_dataset.id,
    )
    ts2 = TimeSeriesWrite(
        name="Test TimeSeries 2 for Datapoints IO",
        external_id="test_timeseries_2_datapoints_io",
        is_step=False,
        is_string=True,
        data_set_id=toolkit_dataset.id,
    )
    retrieved1 = toolkit_client.time_series.retrieve(external_id=ts.external_id)
    retrieved2 = toolkit_client.time_series.retrieve(external_id=ts2.external_id)
    if retrieved1 is None:
        retrieved1 = toolkit_client.time_series.create(ts)
    if retrieved2 is None:
        retrieved2 = toolkit_client.time_series.create(ts2)
    return retrieved1, retrieved2


class TestUploadCommand:
    def test_upload_datapoints(
        self, toolkit_client: ToolkitClient, two_timeseries: tuple[TimeSeries, TimeSeries], tmp_path: Path
    ) -> None:
        ts1, ts2 = two_timeseries
        assert ts1.is_string is False
        assert ts2.is_string is True
        selector = DataPointsFileSelector(
            timestamp_column="timestamp",
            columns=(
                ExternalIdColumn(
                    column="value",
                    external_id=ts1.external_id,
                    dtype="numeric",
                ),
                InternalIdColumn(
                    column="value2",
                    internal_id=ts2.id,
                    dtype="string",
                ),
            ),
        )
        selector.dump_to_file(tmp_path)
        csv_file = tmp_path / f"{selector!s}.{DatapointsIO.KIND}.csv"

        with csv_file.open("w") as f:
            f.write("timestamp,value,value2\n")
            for i in range(10):
                f.write(f"2024-01-01T00:00:{i:02d}Z,{i},no_{i * 10}\n")

        upload_cmd = UploadCommand(silent=True, skip_tracking=True)
        upload_cmd.upload(
            tmp_path,
            toolkit_client,
            deploy_resources=False,
            dry_run=False,
            verbose=True,
        )

        datapoints = toolkit_client.time_series.data.retrieve_arrays(
            external_id=ts1.external_id,
            start=datetime.fromisoformat("2024-01-01T00:00:00Z"),
            end=datetime.fromisoformat("2024-01-01T00:00:10Z"),
        )
        assert len(datapoints) == 10, f"Expected 10 datapoints, got {len(datapoints)}"
        datapoints2 = toolkit_client.time_series.data.retrieve_arrays(
            external_id=ts2.external_id,
            start=datetime.fromisoformat("2024-01-01T00:00:00Z"),
            end=datetime.fromisoformat("2024-01-01T00:00:10Z"),
        )
        assert len(datapoints2) == 10, f"Expected 10 datapoints, got {len(datapoints2)}"



import json
from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any, final

from cognite.client import _version as CogniteSDKVersion
from cognite.client.data_classes.capabilities import (
    Capability,
    FilesAcl,
)
from cognite.client.utils.useful_types import SequenceNotStr
from packaging.requirements import Requirement
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.streamlit_ import (
    Streamlit,
    StreamlitList,
    StreamlitWrite,
    StreamlitWriteList,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotADirectoryError, ToolkitRequiredValueError
from cognite_toolkit._cdf_tk.resource_classes import StreamlitYAML
from cognite_toolkit._cdf_tk.utils import (
    load_yaml_inject_variables,
    safe_read,
)
from cognite_toolkit._cdf_tk.utils.hashing import calculate_hash

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD


@final
class StreamlitCRUD(ResourceCRUD[str, StreamlitWrite, Streamlit, StreamlitWriteList, StreamlitList]):
    folder_name = "streamlit"
    filename_pattern = r".*streamlit$"
    resource_cls = Streamlit
    resource_write_cls = StreamlitWrite
    list_cls = StreamlitList
    list_write_cls = StreamlitWriteList
    kind = "Streamlit"
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD})
    _doc_url = "Files/operation/initFileUpload"
    _metadata_hash_key = "cdf-toolkit-app-hash"
    yaml_cls = StreamlitYAML

    @property
    def display_name(self) -> str:
        return "Streamlit apps"

    @classmethod
    def recommended_packages(cls) -> list[Requirement]:
        return [Requirement("pyodide-http==0.2.1"), Requirement(f"cognite-sdk=={CogniteSDKVersion.__version__}")]

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._source_file_by_external_id: dict[str, Path] = {}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[StreamlitWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        scope: FilesAcl.Scope.All | FilesAcl.Scope.DataSet = FilesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = FilesAcl.Scope.DataSet(list(data_set_ids))

        return FilesAcl(actions, scope)

    @classmethod
    def get_id(cls, item: Streamlit | StreamlitWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("TimeSeries must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        raw_yaml = load_yaml_inject_variables(
            self.safe_read(filepath),
            environment_variables or {},
            original_filepath=filepath,
        )
        raw_list = raw_yaml if isinstance(raw_yaml, list) else [raw_yaml]
        for item in raw_list:
            item_id = self.get_id(item)
            self._source_file_by_external_id[item_id] = filepath
            content = self._as_json_string(item_id, item["entrypoint"])
            item["cogniteToolkitAppHash"] = calculate_hash(content, shorten=True)
        return raw_list

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> StreamlitWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return StreamlitWrite._load(resource)

    def dump_resource(self, resource: Streamlit, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if dumped.get("theme") == "Light" and "theme" not in local:
            dumped.pop("theme")
        return dumped

    @lru_cache
    def _as_json_string(self, external_id: str, entrypoint: str) -> str:
        source_file = self._source_file_by_external_id[external_id]
        app_path = source_file.with_name(external_id)
        if not app_path.exists():
            raise ToolkitNotADirectoryError(f"Streamlit app folder does not exists. Expected: {app_path}")
        requirements_txt = app_path / "requirements.txt"

        if requirements_txt.exists():
            requirements_txt_lines = safe_read(requirements_txt).splitlines()
        else:
            requirements_txt_lines = [str(r) for r in self.recommended_packages()]
        files = {
            py_file.relative_to(app_path).as_posix(): {"content": {"text": safe_read(py_file), "$case": "text"}}
            for py_file in app_path.rglob("*.py")
            if py_file.is_file()
        }

        return json.dumps(
            {
                "entrypoint": entrypoint,
                "files": files,
                "requirements": requirements_txt_lines,
            }
        )

    @staticmethod
    def _missing_recommended_requirements(requirements: list[str]) -> list[str]:
        missing = []
        user_requirements = {Requirement(req).name for req in requirements}
        for recommended in StreamlitCRUD.recommended_packages():
            if recommended.name not in user_requirements:
                missing.append(recommended.name)
        return missing

    def create(self, items: StreamlitWriteList) -> StreamlitList:
        created = StreamlitList([])
        for item in items:
            content = self._as_json_string(item.external_id, item.entrypoint)
            to_create = item.as_file()
            created_file, _ = self.client.files.create(to_create)

            self.client.files.upload_content_bytes(content, item.external_id)
            created.append(Streamlit.from_file(created_file))
        return created

    def retrieve(self, ids: SequenceNotStr[str]) -> StreamlitList:
        files = self.client.files.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)
        return StreamlitList([Streamlit.from_file(file) for file in files])

    def update(self, items: StreamlitWriteList) -> StreamlitList:
        files = []
        for item in items:
            content = self._as_json_string(item.external_id, item.entrypoint)
            to_update = item.as_file()
            self.client.files.upload_content_bytes(content, item.external_id)
            files.append(to_update)

        updated = self.client.files.update(files, mode="replace")
        return StreamlitList([Streamlit.from_file(file) for file in updated])

    def delete(self, ids: SequenceNotStr[str]) -> int:
        self.client.files.delete(external_id=ids)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Streamlit]:
        for file in self.client.files:
            if file.directory == "/streamlit-apps/":
                yield Streamlit.from_file(file)



"""Unit tests for CommandTrackingInfo data class."""

from cognite_toolkit._cdf_tk.data_classes import CommandTrackingInfo


class TestCommandTrackingInfo:
    """Test CommandTrackingInfo functionality."""

    def test_to_dict_empty(self) -> None:
        """Test that empty CommandTrackingInfo returns empty dict."""
        info = CommandTrackingInfo()
        result = info.to_dict()
        assert result == {}

    def test_to_dict_excludes_default_values(self) -> None:
        """Test that empty sets are excluded from to_dict output."""
        info = CommandTrackingInfo()
        info.module_ids.add("module1")
        result = info.to_dict()
        assert "moduleIds" in result
        assert result["moduleIds"] == ["module1"]
        assert "packageIds" not in result
        assert "installedModuleIds" not in result

    def test_to_dict_includes_all_non_empty_fields(self) -> None:
        """Test that all non-empty fields are included with camelCase aliases."""
        info = CommandTrackingInfo(
            project="test_project",
            cluster="test_cluster",
        )
        info.module_ids.add("module1")
        info.package_ids.add("package1")
        info.installed_module_ids.add("installed1")
        info.downloaded_library_ids.add("lib1")

        result = info.to_dict()

        assert result["project"] == "test_project"
        assert result["cluster"] == "test_cluster"
        assert result["moduleIds"] == ["module1"]
        assert result["packageIds"] == ["package1"]
        assert result["installedModuleIds"] == ["installed1"]
        assert result["downloadedLibraryIds"] == ["lib1"]
        # Empty sets should not be present
        assert "downloadedPackageIds" not in result
        assert "downloadedModuleIds" not in result
        assert "installedPackageIds" not in result



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import DatabaseYAML, TableYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_database_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Asset 1"},
        {"Unused field: 'name'", "Missing required field: 'dbName'"},
        id="Missing required field: dbName",
    )
    yield pytest.param(
        {"dbName": "myDB" * 32},
        {"In field dbName string should have at most 32 characters"},
        id="Above maximum character length for dbName",
    )


def invalid_table_test_cases() -> Iterable:
    yield pytest.param(
        {"tableName": "MyTable"}, {"Missing required field: 'dbName'"}, id="Missing required field: dbName"
    )
    yield pytest.param(
        {"dbName": "myDB" * 32, "tableName": "MyTable"},
        {"In field dbName string should have at most 32 characters"},
        id="Above maximum character length for dbName",
    )


class TestRawDatabaseTableYAML:
    @pytest.mark.parametrize("data", list(find_resources("Database")))
    def test_load_valid_database(self, data: dict[str, object]) -> None:
        loaded = DatabaseYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", list(find_resources("Table")))
    def test_load_valid_table(self, data: dict[str, object]) -> None:
        loaded = TableYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_database_test_cases()))
    def test_invalid_database_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, DatabaseYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors

    @pytest.mark.parametrize("data, expected_errors", list(invalid_table_test_cases()))
    def test_invalid_table_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, TableYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import threading
from collections import UserDict
from collections.abc import Iterator
from typing import Any, Generic

from cognite_toolkit._cdf_tk.utils.useful_types import T_ID, T_Value


class ThreadSafeDict(UserDict, Generic[T_ID, T_Value]):
    """A thread-safe dictionary implementation.

    This class extends `UserDict` to provide a dictionary that is safe for concurrent access
    by multiple threads. It uses a threading lock to ensure that operations on the dictionary
    are atomic and do not lead to race conditions.

    It only supports one-thread reading/writing at a time, meaning that if one thread is reading,
    no other thread can read or write until the first thread is done. Todo: Support multiple readers
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._lock = threading.RLock()  # Initialize lock first
        super().__init__(*args, **kwargs)

    def __getitem__(self, key: T_ID) -> T_Value:
        with self._lock:
            return super().__getitem__(key)

    def __setitem__(self, key: T_ID, value: T_Value) -> None:
        with self._lock:
            super().__setitem__(key, value)

    def __delitem__(self, key: T_ID) -> None:
        with self._lock:
            super().__delitem__(key)

    def __len__(self) -> int:
        with self._lock:
            return super().__len__()

    def __contains__(self, key: object) -> bool:
        with self._lock:
            return super().__contains__(key)

    def __iter__(self) -> Iterator[T_ID]:
        with self._lock:
            # Create a copy of keys to avoid iteration issues
            return iter(list(self.data.keys()))

    def clear(self) -> None:
        with self._lock:
            super().clear()

    def copy(self) -> "ThreadSafeDict[T_ID, T_Value]":
        with self._lock:
            new_dict = ThreadSafeDict[T_ID, T_Value]()
            new_dict.data = self.data.copy()
            return new_dict

    def get(self, key: T_ID, default: T_Value | None = None) -> T_Value | None:  # type: ignore[override]
        with self._lock:
            return super().get(key, default)

    def pop(self, key: T_ID, *args: Any) -> T_Value:
        with self._lock:
            return super().pop(key, *args)

    def popitem(self) -> tuple[T_ID, T_Value]:
        with self._lock:
            return super().popitem()

    def setdefault(self, key: T_ID, default: T_Value | None = None) -> T_Value | None:
        with self._lock:
            return super().setdefault(key, default)

    def update(self, *args: Any, **kwargs: Any) -> None:
        with self._lock:
            super().update(*args, **kwargs)

    # For these we override to get a thread-safe operation.
    # keys, values, items in the super class returns views which are not thread-safe
    def keys(self) -> list[T_ID]:  # type: ignore[override]
        with self._lock:
            return list(self.data.keys())

    def values(self) -> list[T_Value]:  # type: ignore[override]
        with self._lock:
            return list(self.data.values())

    def items(self) -> list[tuple[T_ID, T_Value]]:  # type: ignore[override]
        with self._lock:
            return list(self.data.items())

    def __repr__(self) -> str:
        with self._lock:
            return f"{self.__class__.__name__}({dict(self.data)})"

    def __str__(self) -> str:
        with self._lock:
            return str(dict(self.data))



from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Generic, Literal, Protocol, TypeAlias, TypeVar

import httpx
from cognite.client.utils import _json

from cognite_toolkit._cdf_tk.utils.http_client._exception import ToolkitAPIError
from cognite_toolkit._cdf_tk.utils.http_client._tracker import ItemsRequestTracker
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal, PrimitiveType

StatusCode: TypeAlias = int


@dataclass
class HTTPMessage(ABC):
    """Base class for HTTP messages (requests and responses)"""

    def dump(self) -> dict[str, JsonVal]:
        """Dumps the message to a JSON serializable dictionary.

        Returns:
            dict[str, JsonVal]: The message as a dictionary.
        """
        # We avoid using the asdict function as we know we have a shallow structure,
        # and this roughly ~10x faster.
        output = self.__dict__.copy()
        output["type"] = type(self).__name__
        return output


@dataclass
class ErrorDetails:
    """This is the expected structure of error details in the CDF API"""

    code: int
    message: str
    missing: list[JsonVal] | None = None
    duplicated: list[JsonVal] | None = None
    is_auto_retryable: bool | None = None

    @classmethod
    def from_response(cls, response: httpx.Response) -> "ErrorDetails":
        try:
            error_data = response.json()["error"]
            if not isinstance(error_data, dict):
                return cls(code=response.status_code, message=str(error_data))
            return cls(
                code=error_data["code"],
                message=error_data["message"],
                missing=error_data.get("missing"),
                duplicated=error_data.get("duplicated"),
                is_auto_retryable=error_data.get("isAutoRetryable"),
            )
        except (ValueError, KeyError):
            # Fallback if response is not JSON or does not have expected structure
            return cls(code=response.status_code, message=response.text)

    def dump(self) -> dict[str, JsonVal]:
        output: dict[str, JsonVal] = {
            "code": self.code,
            "message": self.message,
        }
        if self.missing is not None:
            output["missing"] = self.missing
        if self.duplicated is not None:
            output["duplicated"] = self.duplicated
        if self.is_auto_retryable is not None:
            output["isAutoRetryable"] = self.is_auto_retryable
        return output


@dataclass
class FailedRequestMessage(HTTPMessage):
    error: str


@dataclass
class ResponseMessage(HTTPMessage):
    status_code: StatusCode


@dataclass
class RequestMessage(HTTPMessage):
    """Base class for HTTP request messages"""

    endpoint_url: str
    method: Literal["GET", "POST", "PATCH", "DELETE"]
    connect_attempt: int = 0
    read_attempt: int = 0
    status_attempt: int = 0
    api_version: str | None = None
    content_type: str = "application/json"
    accept: str = "application/json"

    @property
    def total_attempts(self) -> int:
        return self.connect_attempt + self.read_attempt + self.status_attempt

    @abstractmethod
    def create_success_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        raise NotImplementedError()

    @abstractmethod
    def create_failure_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        raise NotImplementedError()

    @abstractmethod
    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        raise NotImplementedError()


@dataclass
class SuccessResponse(ResponseMessage):
    body: str


@dataclass
class FailedResponse(ResponseMessage):
    body: str
    error: ErrorDetails

    def dump(self) -> dict[str, JsonVal]:
        output = super().dump()
        output["error"] = self.error.dump()
        return output


@dataclass
class SimpleRequest(RequestMessage):
    """Base class for requests with a simple success/fail response structure"""

    @classmethod
    def create_success_response(cls, response: httpx.Response) -> Sequence[ResponseMessage]:
        return [SuccessResponse(status_code=response.status_code, body=response.text)]

    @classmethod
    def create_failure_response(cls, response: httpx.Response) -> Sequence[HTTPMessage]:
        return [
            FailedResponse(
                status_code=response.status_code, error=ErrorDetails.from_response(response), body=response.text
            )
        ]

    @classmethod
    def create_failed_request(cls, error_message: str) -> Sequence[HTTPMessage]:
        return [FailedRequestMessage(error=error_message)]


@dataclass
class BodyRequest(RequestMessage, ABC):
    """Base class for HTTP request messages with a body"""

    @abstractmethod
    def data(self) -> str:
        raise NotImplementedError()


@dataclass
class ParamRequest(SimpleRequest):
    """Base class for HTTP request messages with query parameters"""

    parameters: dict[str, PrimitiveType] | None = None


@dataclass
class SimpleBodyRequest(SimpleRequest, BodyRequest):
    body_content: dict[str, JsonVal] = field(default_factory=dict)

    def data(self) -> str:
        return _dump_body(self.body_content)


@dataclass
class DataBodyRequest(SimpleRequest):
    data_content: bytes = b""

    def data(self) -> bytes:
        return self.data_content

    def dump(self) -> dict[str, JsonVal]:
        output = super().dump()
        # We cannot serialize bytes, so we indicate its presence instead
        output["data_content"] = "<bytes>"
        return output


T_COVARIANT_ID = TypeVar("T_COVARIANT_ID", covariant=True)


@dataclass
class ItemMessage(Generic[T_COVARIANT_ID], ABC):
    """Base class for message related to a specific item identified by an ID"""

    ids: list[T_COVARIANT_ID] = field(default_factory=list)


@dataclass
class SuccessResponseItems(ItemMessage[T_COVARIANT_ID], SuccessResponse): ...


@dataclass
class FailedResponseItems(ItemMessage[T_COVARIANT_ID], FailedResponse): ...


@dataclass
class FailedRequestItems(ItemMessage[T_COVARIANT_ID], FailedRequestMessage): ...


class RequestItem(Generic[T_COVARIANT_ID], Protocol):
    def dump(self) -> JsonVal: ...

    def as_id(self) -> T_COVARIANT_ID: ...


@dataclass
class ItemsRequest(Generic[T_COVARIANT_ID], BodyRequest):
    """Requests message for endpoints that accept multiple items in a single request.

    This class provides functionality to split large requests into smaller ones, handle responses for each item,
    and manage errors effectively.

    Attributes:
        items (list[T_RequestItem]): The list of items to be sent in the request body.
        extra_body_fields (dict[str, JsonVal]): Additional fields to include in the request body
        max_failures_before_abort (int): The maximum number of failed split requests before aborting further splits.

    """

    items: list[RequestItem[T_COVARIANT_ID]] = field(default_factory=list)
    extra_body_fields: dict[str, JsonVal] = field(default_factory=dict)
    max_failures_before_abort: int = 50
    tracker: ItemsRequestTracker | None = field(default=None, init=False)

    def dump(self) -> dict[str, JsonVal]:
        """Dumps the message to a JSON serializable dictionary.

        This override removes the 'as_id' attribute as it is not serializable.

        Returns:
            dict[str, JsonVal]: The message as a dictionary.
        """
        output = super().dump()
        output["items"] = self.dump_items()
        if self.tracker is not None:
            # We cannot serialize the tracker
            del output["tracker"]
        return output

    def dump_items(self) -> list[JsonVal]:
        """Dumps the items to a list of JSON serializable dictionaries.

        Returns:
            list[JsonVal]: The items as a list of dictionaries.
        """
        return [item.dump() for item in self.items]

    def body(self) -> dict[str, JsonVal]:
        if self.extra_body_fields:
            return {"items": self.dump_items(), **self.extra_body_fields}
        return {"items": self.dump_items()}

    def data(self) -> str:
        return _dump_body(self.body())

    def split(self, status_attempts: int) -> "list[ItemsRequest]":
        """Splits the request into two smaller requests.

        This is useful for retrying requests that fail due to size limits or timeouts.

        Args:
            status_attempts: The number of status attempts to set for the new requests. This is used when the
                request failed with a 5xx status code and we want to track the number of attempts. For 4xx errors,
                there is at least one item causing the error, so we do not increment the status attempts, but
                instead essentially do a binary search to find the problematic item(s).

        Returns:
            A list containing two new ItemsRequest instances, each with half of the original items.

        """
        mid = len(self.items) // 2
        if mid == 0:
            return [self]
        tracker = self.tracker or ItemsRequestTracker(self.max_failures_before_abort)
        tracker.register_failure()
        first_half = ItemsRequest[T_COVARIANT_ID](
            endpoint_url=self.endpoint_url,
            method=self.method,
            items=self.items[:mid],
            extra_body_fields=self.extra_body_fields,
            connect_attempt=self.connect_attempt,
            read_attempt=self.read_attempt,
            status_attempt=status_attempts,
        )
        first_half.tracker = tracker
        second_half = ItemsRequest[T_COVARIANT_ID](
            endpoint_url=self.endpoint_url,
            method=self.method,
            items=self.items[mid:],
            extra_body_fields=self.extra_body_fields,
            connect_attempt=self.connect_attempt,
            read_attempt=self.read_attempt,
            status_attempt=status_attempts,
        )
        second_half.tracker = tracker
        return [first_half, second_half]

    def create_success_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        ids = [item.as_id() for item in self.items]
        return [SuccessResponseItems(status_code=response.status_code, ids=ids, body=response.text)]

    def create_failure_response(self, response: httpx.Response) -> Sequence[HTTPMessage]:
        error = ErrorDetails.from_response(response)
        ids = [item.as_id() for item in self.items]
        return [FailedResponseItems(status_code=response.status_code, ids=ids, error=error, body=response.text)]

    def create_failed_request(self, error_message: str) -> Sequence[HTTPMessage]:
        ids = [item.as_id() for item in self.items]
        return [FailedRequestItems(ids=ids, error=error_message)]


class ResponseList(UserList[ResponseMessage | FailedRequestMessage]):
    def __init__(self, collection: Sequence[ResponseMessage | FailedRequestMessage] | None = None) -> None:
        super().__init__(collection or [])

    def raise_for_status(self) -> None:
        """Raises an exception if any response in the list indicates a failure."""
        failed_responses = [resp for resp in self.data if isinstance(resp, FailedResponse)]
        failed_requests = [resp for resp in self.data if isinstance(resp, FailedRequestMessage)]
        if not failed_responses and not failed_requests:
            return
        error_messages = "; ".join(f"Status {err.status_code}: {err.error}" for err in failed_responses)
        if failed_requests:
            if error_messages:
                error_messages += "; "
            error_messages += "; ".join(f"Request error: {err.error}" for err in failed_requests)
        raise ToolkitAPIError(f"One or more requests failed: {error_messages}")

    def get_first_body(self) -> dict[str, JsonVal]:
        """Returns the body of the first successful response in the list.

        Raises:
            ValueError: If there are no successful responses with a body.

        Returns:
            dict[str, JsonVal]: The body of the first successful response.
        """
        for resp in self.data:
            if isinstance(resp, SuccessResponse) and resp.body is not None:
                return _json.loads(resp.body)
        raise ValueError("No successful responses with a body found.")


def _dump_body(body: dict[str, JsonVal]) -> str:
    try:
        return _json.dumps(body, allow_nan=False)
    except ValueError as e:
        # A lot of work to give a more human friendly error message when nans and infs are present:
        msg = "Out of range float values are not JSON compliant"
        if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
            raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                e.__traceback__
            ) from None
        raise
